# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Classes and routines for creating widgets from `Gst.Element`s."""
import math
import os
import re
from gettext import gettext as _

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GES
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GstController
from gi.repository import Gtk
from gi.repository import Pango

from pitivi.configure import get_ui_dir
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import is_valid_file
from pitivi.utils.timeline import Zoomable
from pitivi.utils.ui import beautify_length
from pitivi.utils.ui import disable_scroll
from pitivi.utils.ui import SPACING
from pitivi.utils.ui import time_to_string


ZOOM_SLIDER_PADDING = SPACING * 4 / 5


class DynamicWidget(Loggable):
    """Abstract widget providing a way to get, set and observe properties."""

    def __init__(self, default):
        super().__init__()
        self.default = default

    def connect_value_changed(self, callback, *args):
        raise NotImplementedError

    def set_widget_value(self, value):
        raise NotImplementedError

    def get_widget_value(self):
        raise NotImplementedError

    def get_widget_default(self):
        return self.default

    def set_widget_to_default(self):
        if self.default is not None:
            self.set_widget_value(self.default)


class DefaultWidget(Gtk.Label):
    """When all hope fails...."""

    def __init__(self):
        Gtk.Label.__init__(self, _("Implement Me"))
        self.props.halign = Gtk.Align.START

    def connect_value_changed(self, callback, *args):
        pass

    def set_widget_value(self, value):
        pass

    def get_widget_value(self):
        pass

    def set_widget_to_default(self):
        pass


class TextWidget(Gtk.Box, DynamicWidget):
    """Widget for entering text.

    A Gtk.Entry which emits a "value-changed" signal only when its input is
    valid (matches the provided regex). If the input is invalid, a warning
    icon is displayed.

    You can also connect to the "activate" signal if you don't want to watch
    for live changes, but it will only be emitted if the input is valid when
    the user presses Enter.
    """

    __gtype_name__ = 'TextWidget'
    __gsignals__ = {
        "value-changed": (GObject.SignalFlags.RUN_LAST, None, (),),
        "activate": (GObject.SignalFlags.RUN_LAST, None, (),)
    }

    def __init__(self, matches=None, choices=None, default=None, combobox=False, widget=None):
        if not default:
            # In the case of text widgets, a blank default is an empty string
            default = ""

        Gtk.Box.__init__(self)
        DynamicWidget.__init__(self, default)

        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.set_border_width(0)
        self.set_spacing(0)
        if widget is None:
            if choices:
                self.combo = Gtk.ComboBoxText.new_with_entry()
                self.text = self.combo.get_child()
                self.combo.show()
                self.pack_start(self.combo, expand=False, fill=False, padding=0)
                for choice in choices:
                    self.combo.append_text(choice)
            elif combobox:
                self.combo = Gtk.ComboBox.new_with_entry()
                self.text = self.combo.get_child()
                self.combo.show()
                self.pack_start(self.combo, expand=False, fill=False, padding=0)
            else:
                self.text = Gtk.Entry()
                self.text.show()
                self.pack_start(self.text, expand=False, fill=False, padding=0)
        else:
            self.text = widget

        self.matches = None
        self.last_valid = None
        self.valid = False
        self.send_signal = True
        self.text.connect("changed", self.__text_changed_cb)
        self.text.connect("activate", self.__activate_cb)
        if matches:
            if isinstance(matches, str):
                self.matches = re.compile(matches)
            else:
                self.matches = matches
            self.__text_changed_cb(None)

    def connect_value_changed(self, callback, *args):
        return self.connect("value-changed", callback, *args)

    def set_widget_value(self, value, send_signal=True):
        self.send_signal = send_signal
        self.text.set_text(value)

    def get_widget_value(self):
        if self.matches:
            return self.last_valid
        return self.text.get_text()

    def add_choices(self, choices):
        for choice in choices:
            self.combo.append_text(choice)

    def __text_changed_cb(self, unused_widget):
        text = self.text.get_text()
        if self.matches:
            if self._filter(text):
                self.last_valid = text
                if self.send_signal:
                    self.emit("value-changed")
                if not self.valid:
                    self.text.set_icon_from_icon_name(1, None)
                self.valid = True
            else:
                if self.valid:
                    self.text.set_icon_from_icon_name(1, "dialog-warning")
                self.valid = False
        elif self.send_signal:
            self.emit("value-changed")

        self.send_signal = True

    def __activate_cb(self, unused_widget):
        if self.matches and self.send_signal:
            self.emit("activate")

    def _filter(self, text):
        match = self.matches.match(text)
        if match is not None:
            return True
        return False

    def set_width_chars(self, width):
        """Allows setting the width of the text entry widget for compactness."""
        self.text.set_width_chars(width)


class NumericWidget(Gtk.Box, DynamicWidget):
    """Widget for entering a number.

    Contains both a Gtk.Scale and a Gtk.SpinButton for adjusting the value.
    The SpinButton is always displayed, while the Scale only appears if both
    lower and upper bounds are defined.

    Args:
        upper (Optional[int]): The upper limit for this widget.
        lower (Optional[int]): The lower limit for this widget.
    """

    def __init__(self, upper=None, lower=None, default=None, adjustment=None, width_chars=None):
        Gtk.Box.__init__(self)
        DynamicWidget.__init__(self, default)

        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.set_spacing(SPACING)
        self._type = None
        self.spinner = None
        self.handler_id = None

        if adjustment:
            self.adjustment = adjustment
            return
        reasonable_limit = 5000
        with_slider = (lower is not None and lower > -reasonable_limit and
                       upper is not None and upper < reasonable_limit)
        self.adjustment = Gtk.Adjustment()
        if upper is None:
            upper = GLib.MAXINT32
        if lower is None:
            lower = GLib.MININT32
        self.adjustment.props.lower = lower
        self.adjustment.props.upper = upper

        self.spinner = Gtk.SpinButton(adjustment=self.adjustment)
        if width_chars:
            self.spinner.props.width_chars = width_chars
        self.pack_start(self.spinner, expand=False, fill=False, padding=0)
        self.spinner.show()

        if with_slider:
            self.slider = Gtk.Scale.new(
                Gtk.Orientation.HORIZONTAL, self.adjustment)
            self.pack_start(self.slider, expand=False, fill=False, padding=0)
            self.slider.show()
            self.slider.set_size_request(width=100, height=-1)
            self.slider.props.draw_value = False
            # Abuse GTK3's progressbar "fill level" feature to provide
            # a visual indication of the default value on property sliders.
            if default is not None:
                self.slider.set_restrict_to_fill_level(False)
                self.slider.set_fill_level(float(default))
                self.slider.set_show_fill_level(True)

    def block_signals(self):
        if self.handler_id:
            self.adjustment.handler_block(self.handler_id)

    def unblock_signals(self):
        if self.handler_id:
            self.adjustment.handler_unblock(self.handler_id)

    def connect_value_changed(self, callback, *args):
        self.handler_id = self.adjustment.connect("value-changed", callback, *args)

    def get_widget_value(self):
        if self._type:
            return self._type(self.adjustment.get_value())

        return self.adjustment.get_value()

    def set_widget_value(self, value):
        type_ = type(value)
        if self._type is None:
            self._type = type_

        if type_ == int:
            step = 1.0
            page = 10.0
        elif type_ == float:
            step = 0.01
            page = 0.1
            if self.spinner:
                self.spinner.props.digits = 2
        else:
            raise Exception('Unsupported property type: %s' % type_)
        lower = min(self.adjustment.props.lower, value)
        upper = max(self.adjustment.props.upper, value)
        self.adjustment.configure(value, lower, upper, step, page, 0)


class TimeWidget(TextWidget, DynamicWidget):
    """Widget for entering a time value.

    Accepts timecode formats or a frame number (integer).
    """

    # The "frame number" match rule is ^([0-9]+)$ (with a + to require 1 digit)
    # The "timecode" rule is ^([0-9]:[0-5][0-9]:[0-5][0-9])\.[0-9][0-9][0-9]$"
    # Combining the two, we get:
    VALID_REGEX = re.compile(
        r"^([0-9]+)$|^([0-9]:)?([0-5][0-9]:[0-5][0-9])\.[0-9][0-9][0-9]$")

    __gtype_name__ = 'TimeWidget'

    def __init__(self, default=None):
        DynamicWidget.__init__(self, default)
        TextWidget.__init__(self, self.VALID_REGEX)
        TextWidget.set_width_chars(self, 10)
        self._framerate = None
        self.text.connect("focus-out-event", self._focus_out_cb)

    def get_widget_value(self):
        timecode = TextWidget.get_widget_value(self)

        if ":" in timecode:
            parts = timecode.split(":")
            if len(parts) == 2:
                hh = 0
                mm, end = parts
            else:
                hh, mm, end = parts
            ss, millis = end.split(".")
            nanosecs = int(hh) * 3.6 * 10e12 \
                + int(mm) * 6 * 10e10 \
                + int(ss) * 10e9 \
                + int(millis) * 10e6
            nanosecs = nanosecs / 10  # Compensate the 10 factor of e notation
        else:
            # We were given a frame number. Convert from the project framerate.
            frame_no = int(timecode)
            nanosecs = frame_no / float(self._framerate) * Gst.SECOND
        return int(nanosecs)

    def set_widget_value(self, timeNanos, send_signal=True):
        self.default = timeNanos
        timecode = time_to_string(timeNanos)
        if timecode.startswith("0:"):
            timecode = timecode[2:]
        TextWidget.set_widget_value(self, timecode, send_signal=send_signal)

    def _focus_out_cb(self, widget, event):
        """Reset the text to display the current position of the playhead."""
        if self.default is not None:
            self.set_widget_value(self.default)

    def connect_activate_event(self, activate_cb):
        return self.connect("activate", activate_cb)

    def connect_focus_events(self, focus_in_cb, focus_out_cb):
        focus_in_handler_id = self.text.connect("focus-in-event", focus_in_cb)
        focus_out_handler_id = self.text.connect("focus-out-event", focus_out_cb)
        return [focus_in_handler_id, focus_out_handler_id]

    def set_framerate(self, framerate):
        self._framerate = framerate


class FractionWidget(TextWidget, DynamicWidget):
    """Widget for entering a fraction."""

    fraction_regex = re.compile(
        r"^([0-9]*(\.[0-9]+)?)(([:/][0-9]*(\.[0-9]+)?)|M)?$")
    __gtype_name__ = 'FractionWidget'

    def __init__(self, presets=None, default=None):
        DynamicWidget.__init__(self, default)

        flow = float("-Infinity")
        fhigh = float("Infinity")
        choices = []
        if presets:
            for preset in presets:
                if isinstance(preset, str):
                    strval = preset
                    preset = self._parse_text(preset)
                else:
                    strval = "%g:%g" % (preset.num, preset.denom)
                if flow <= float(preset) <= fhigh:
                    choices.append(strval)
        self.low = flow
        self.high = fhigh
        TextWidget.__init__(self, self.fraction_regex, choices)

    def _filter(self, text):
        if TextWidget._filter(self, text):
            value = self._parse_text(text)
            if self.low <= float(value) and float(value) <= self.high:
                return True
        return False

    def add_presets(self, presets):
        choices = []
        for preset in presets:
            if isinstance(preset, str):
                strval = preset
                preset = self._parse_text(preset)
            else:
                strval = "%g:%g" % (preset.num, preset.denom)
            if self.low <= float(preset) <= self.high:
                choices.append(strval)

        self.add_choices(choices)

    def set_widget_value(self, value):
        if isinstance(value, str):
            value = self._parse_text(value)
        elif not hasattr(value, "denom"):
            value = Gst.Fraction(value)
        if value.denom == 1001:
            text = "%dM" % (value.num / 1000)
        else:
            text = "%d:%d" % (value.num, value.denom)

        self.text.set_text(text)

    def get_widget_value(self):
        if self.last_valid:
            return self._parse_text(self.last_valid)
        return Gst.Fraction(1, 1)

    @classmethod
    def _parse_text(cls, text):
        match = cls.fraction_regex.match(text)
        groups = match.groups()
        num = 1.0
        denom = 1.0
        if groups[0]:
            num = float(groups[0])
        if groups[2]:
            if groups[2] == "M":
                num = num * 1000
                denom = 1001
            elif groups[2][1:]:
                denom = float(groups[2][1:])
        return Gst.Fraction(num, denom)


class ToggleWidget(Gtk.Box, DynamicWidget):
    """Widget for entering an on/off value."""

    def __init__(self, default=None, switch_button=None):
        Gtk.Box.__init__(self)
        DynamicWidget.__init__(self, default)
        self.props.valign = Gtk.Align.CENTER
        if switch_button is None:
            self.switch_button = Gtk.Switch()
            self.pack_start(self.switch_button, expand=False, fill=False, padding=0)
            self.switch_button.show()
        else:
            self.switch_button = switch_button
            self.set_widget_to_default()

    def connect_value_changed(self, callback, *args):
        def callback_wrapper(switch_button, unused_state):
            callback(switch_button, *args)

        self.switch_button.connect("state-set", callback_wrapper)

    def set_widget_value(self, value):
        self.switch_button.set_active(value)

    def get_widget_value(self):
        return self.switch_button.get_active()


class ChoiceWidget(Gtk.Box, DynamicWidget):
    """Widget for making a choice between a list of named values."""

    def __init__(self, choices, default=None):
        Gtk.Box.__init__(self)
        DynamicWidget.__init__(self, default)
        self.choices = None
        self.values = None
        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.contents = Gtk.ComboBoxText()
        self.pack_start(self.contents, expand=False, fill=False, padding=0)
        self.set_choices(choices)
        self.contents.show()
        cell = self.contents.get_cells()[0]
        cell.props.ellipsize = Pango.EllipsizeMode.END

    def connect_value_changed(self, callback, *args):
        return self.contents.connect("changed", callback, *args)

    def set_widget_value(self, value):
        try:
            self.contents.set_active(self.values.index(value))
        except ValueError as e:
            raise ValueError("%r not in %r" % (value, self.values)) from e

    def get_widget_value(self):
        return self.values[self.contents.get_active()]

    def set_choices(self, choices):
        self.choices = [choice[0] for choice in choices]
        self.values = [choice[1] for choice in choices]
        model = Gtk.ListStore(str)
        self.contents.set_model(model)
        for choice, unused_value in choices:
            self.contents.append_text(_(choice))
        if len(choices) <= 1:
            self.contents.set_sensitive(False)
        else:
            self.contents.set_sensitive(True)


class PathWidget(Gtk.FileChooserButton, DynamicWidget):
    """Widget for entering a path."""

    __gtype_name__ = 'PathWidget'

    __gsignals__ = {
        "value-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, action=Gtk.FileChooserAction.OPEN, default=None):
        DynamicWidget.__init__(self, default)
        self.dialog = Gtk.FileChooserDialog(action=action)
        self.dialog.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        self.dialog.set_default_response(Gtk.ResponseType.OK)
        Gtk.FileChooserButton.__init__(self, dialog=self.dialog)
        self.dialog.connect("response", self._response_cb)
        self.uri = ""

    def connect_value_changed(self, callback, *args):
        return self.connect("value-changed", callback, *args)

    def set_widget_value(self, value):
        self.set_uri(value)
        self.uri = value

    def get_widget_value(self):
        return self.uri

    def _response_cb(self, unused_dialog, response):
        if response == Gtk.ResponseType.CLOSE:
            self.uri = self.get_uri()
            self.emit("value-changed")
            self.dialog.hide()


class ColorWidget(Gtk.ColorButton, DynamicWidget):

    def __init__(self, default=None):
        Gtk.ColorButton.__init__(self)
        DynamicWidget.__init__(self, default)

    def connect_value_changed(self, callback, *args):
        self.connect("color-set", callback, *args)

    def set_widget_value(self, value):
        self.set_rgba(value)

    def get_widget_value(self):
        return self.get_rgba()


class FontWidget(Gtk.FontButton, DynamicWidget):

    def __init__(self, default=None):
        Gtk.FontButton.__init__(self)
        DynamicWidget.__init__(self, default)
        self.set_use_font(True)

    def connect_value_changed(self, callback, *args):
        self.connect("font-set", callback, *args)

    def set_widget_value(self, font_name):
        self.set_font_name(font_name)

    def get_widget_value(self):
        return self.get_font_name()


class InputValidationWidget(Gtk.Box, DynamicWidget):
    """Widget for validating the input of another widget.

    It shows a warning sign if the input is not valid and rolls back to
    the default widget value (which should always be valid).

    Args:
        widget (DynamicWidget): widget whose input needs validation.
        validation_function (function): function which receives the input of the
            widget and returns True iff the input is valid.
    """

    def __init__(self, widget, validation_function):
        Gtk.Box.__init__(self)
        DynamicWidget.__init__(self, widget.default)
        self._widget = widget
        self._validation_function = validation_function
        self._warning_sign = Gtk.Image.new_from_icon_name("dialog-warning", 3)

        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.pack_start(self._widget, expand=False, fill=False, padding=0)
        self.pack_start(self._warning_sign, expand=False, fill=False, padding=SPACING)
        self._warning_sign.set_no_show_all(True)

        self._widget.connect_value_changed(self._widget_value_changed_cb)

    def connect_value_changed(self, callback, *args):
        return self._widget.connect_value_changed(callback, args)

    def set_widget_value(self, value):
        self._widget.set_widget_value(value)

    def get_widget_value(self):
        value = self._widget.get_widget_value()
        if self._validation_function(value):
            return value
        return self.get_widget_default()

    def _widget_value_changed_cb(self, unused_widget):
        value = self._widget.get_widget_value()
        if self._validation_function(value):
            self._warning_sign.hide()
        else:
            self._warning_sign.show()


def make_widget_wrapper(prop, widget):
    """Creates a wrapper child of DynamicWidget for @widget."""
    # Respect Object hierarchy here
    if isinstance(widget, Gtk.SpinButton):
        widget_adjustment = widget.get_adjustment()
        widget_lower = widget_adjustment.props.lower
        widget_upper = widget_adjustment.props.upper
        return NumericWidget(upper=widget_upper, lower=widget_lower, adjustment=widget_adjustment, default=prop.default_value)
    elif isinstance(widget, Gtk.Entry):
        return TextWidget(widget=widget)
    elif isinstance(widget, Gtk.Range):
        widget_adjustment = widget.get_adjustment()
        widget_lower = widget_adjustment.props.lower
        widget_upper = widget_adjustment.props.upper
        return NumericWidget(upper=widget_upper, lower=widget_lower, adjustment=widget_adjustment, default=prop.default_value)
    elif isinstance(widget, Gtk.Switch):
        return ToggleWidget(prop.default_value, widget)
    else:
        Loggable().fixme("%s has not been wrapped into a Dynamic Widget", widget)
        return None


class GstElementSettingsWidget(Gtk.Box, Loggable):
    """Widget to modify the properties of a Gst.Element.

    Can be used to configure an effect or an encoder, etc.

    Args:
        controllable (bool): Whether the properties being controlled by
            keyframes is allowed.
    """

    # Dictionary that maps tuples of (element_name, property_name) to a
    # validation function.
    INPUT_VALIDATION_FUNCTIONS = {
        ("x264enc", "multipass-cache-file"): is_valid_file
    }

    # Dictionary that references the GstCaps field to expose in the UI
    # for a well known set of elements.
    CAP_FIELDS_TO_EXPOSE = {
        "x264enc": {"profile": Gst.ValueList(["high", "main", "baseline"])}
    }

    def __init__(self, element, props_to_ignore=("name",), controllable=True):
        Gtk.Box.__init__(self)
        Loggable.__init__(self)
        self.element = element
        self.ignore = props_to_ignore
        self.properties = {}
        # Maps caps fields to the corresponding widgets.
        self.caps_widgets = {}
        self.__controllable = controllable
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.__bindings_by_keyframe_button = {}
        self.__widgets_by_keyframe_button = {}
        self.__widgets_by_reset_button = {}
        self._unhandled_properties = []
        self.uncontrolled_properties = {}
        self.updating_property = False

    def deactivate_keyframe_toggle_buttons(self):
        """Makes sure the keyframe togglebuttons are deactivated."""
        self.log("Deactivating all keyframe toggle buttons")
        for keyframe_button in self.__widgets_by_keyframe_button:
            if keyframe_button.get_active():
                # Deactivate the button. The only effect should be that
                # the keyframe curve will control again the default property.
                keyframe_button.set_active(False)
                # There can be only one active keyframes button.
                break

    def show_widget(self, widget):
        self.pack_start(widget, True, True, 0)
        disable_scroll(widget)
        self.show_all()

    def map_builder(self, builder):
        """Maps the GStreamer element's properties to corresponding widgets in @builder.

        Prop control widgets should be named "element_name::prop_name", where:
        - element_name is the gstreamer element (ex: the "alpha" effect)
        - prop_name is the name of one of a particular property of the element
        If present, a reset button corresponding to the property will be used
        (the button must be named similarly, with "::reset" after the prop name)
        A button named reset_all_button can also be provided and will be used as
        a fallback for each property without an individual reset button.
        Similarly, the keyframe control button corresponding to the property (if controllable)
        can be used whose name is to be "element_name::prop_name::keyframe".
        """
        reset_all_button = builder.get_object("reset_all_button")
        for prop in self._get_properties():
            widget_name = prop.owner_type.name + "::" + prop.name
            widget = builder.get_object(widget_name)
            if widget is None:
                self._unhandled_properties.append(prop)
                self.warning("No custom widget found for %s property \"%s\"" %
                             (prop.owner_type.name, prop.name))
            else:
                reset_name = widget_name + "::" + "reset"
                reset_widget = builder.get_object(reset_name)
                if not reset_widget:
                    # If reset_all_button is not found, it will be None
                    reset_widget = reset_all_button
                keyframe_name = widget_name + "::" + "keyframe"
                keyframe_widget = builder.get_object(keyframe_name)
                self.add_property_widget(prop, widget, reset_widget, keyframe_widget)

    def add_property_widget(self, prop, widget, to_default_btn=None, keyframe_btn=None):
        """Connects an element property to a GTK Widget.

        Optionally, a reset button widget can also be provided.
        Unless you want to connect each widget individually, you should be using
        the "map_builder" method instead.
        """
        if isinstance(widget, DynamicWidget):
            # if the widget is already a DynamicWidget we use it as is
            dynamic_widget = widget
        else:
            # if the widget is not dynamic we try to create a wrapper around it
            # so we can control it with the standardized DynamicWidget API
            dynamic_widget = make_widget_wrapper(prop, widget)

        if dynamic_widget:
            self.properties[prop] = dynamic_widget

            self.element.connect("notify::" + prop.name, self._property_changed_cb,
                                 dynamic_widget)
            # The "reset to default" button associated with this property
            if isinstance(to_default_btn, Gtk.Button):
                self.__widgets_by_reset_button[to_default_btn] = widget
                to_default_btn.connect("clicked", self.__reset_to_default_clicked_cb, dynamic_widget, keyframe_btn)
            elif to_default_btn is not None:
                self.warning("to_default_btn should be Gtk.Button or None, got %s", to_default_btn)

            # The "keyframe toggle" button associated with this property
            if not isinstance(widget, (ToggleWidget, ChoiceWidget)):
                res, element, unused_pspec = self.element.lookup_child(prop.name)
                assert res
                binding = GstController.DirectControlBinding.new(
                    element, prop.name,
                    GstController.InterpolationControlSource())
                if binding.pspec:
                    # The prop can be controlled (keyframed).
                    if isinstance(keyframe_btn, Gtk.ToggleButton):
                        keyframe_btn.connect("toggled", self.__keyframes_toggled_cb, prop)
                        self.__widgets_by_keyframe_button[keyframe_btn] = widget
                        prop_binding = self.element.get_control_binding(prop.name)
                        self.__bindings_by_keyframe_button[keyframe_btn] = prop_binding
                        self.__display_controlled(keyframe_btn, bool(prop_binding))
                    elif keyframe_btn is not None:
                        self.warning("keyframe_btn should be Gtk.ToggleButton or None, got %s", to_default_btn)
        else:
            # If we add a non-standard widget, the creator of the widget is
            # responsible for handling its behaviour "by hand"
            self.info("Can not wrap widget %s for property %s" % (widget, prop))
            # We still keep a ref to that widget, "just in case"
            self.uncontrolled_properties[prop] = widget

        if hasattr(prop, "blurb"):
            widget.set_tooltip_text(prop.blurb)

    def _get_properties(self):
        if isinstance(self.element, GES.BaseEffect):
            props = self.element.list_children_properties()
        else:
            props = GObject.list_properties(self.element)
        return [prop for prop in props if prop.name not in self.ignore]

    def __add_widget_to_grid(self, grid, nick, widget, y):
        if isinstance(widget, ToggleWidget):
            widget.set_label(nick)
            grid.attach(widget, 0, y, 2, 1)
        else:
            text = _("%(preference_label)s:") % {"preference_label": nick}
            label = Gtk.Label(label=text)
            label.props.yalign = 0.5
            grid.attach(label, 0, y, 1, 1)
            grid.attach(widget, 1, y, 1, 1)

    def add_widgets(self, create_property_widget_func, values=None, with_reset_button=False, caps_values=None):
        """Prepares a Gtk.Grid containing the property widgets of an element.

        Each property is on a separate row.
        A row is typically a label followed by the widget and a reset button.

        If there are no properties, returns a "No properties" label.

        Args:
            create_property_widget_func (function): The function that creates
                the widget for an effect property.
            values (dict): The current values of the element props, by name.
                If empty, the default values will be used.
            with_reset_button (bool): Whether to show a reset button for each
                property.
            caps_values (Optional[dict]): Map of caps fields to their values.
        """
        values = values or {}
        self.info("element: %s, use values: %s", self.element, values)
        self.properties.clear()
        self.__bindings_by_keyframe_button = {}
        self.__widgets_by_keyframe_button = {}
        is_effect = isinstance(self.element, GES.Effect)
        if is_effect:
            props = [prop for prop in self.element.list_children_properties()
                     if prop.name not in self.ignore]
        else:
            props = [prop for prop in GObject.list_properties(self.element)
                     if prop.name not in self.ignore]
        if not props:
            widget = Gtk.Label(label=_("No properties."))
            self.pack_start(widget, expand=False, fill=False, padding=0)
            widget.show()
            return

        grid = Gtk.Grid()
        grid.props.row_spacing = SPACING
        grid.props.column_spacing = SPACING
        grid.props.border_width = SPACING

        element_name = None
        if isinstance(self.element, Gst.Element):
            element_name = self.element.get_factory().get_name()

        src_caps_fields = self.CAP_FIELDS_TO_EXPOSE.get(element_name)
        y = 0
        if src_caps_fields:
            srccaps = self.element.get_static_pad('src').get_pad_template().caps

            for field, prefered_value in src_caps_fields.items():
                gvalue = srccaps[0][field]
                if isinstance(gvalue, Gst.ValueList) and isinstance(prefered_value, Gst.ValueList):
                    prefered_value = Gst.ValueList([v for v in prefered_value if v in gvalue])
                    gvalue = Gst.ValueList.merge(prefered_value, gvalue)

                widget = self._make_widget_from_gvalue(gvalue, prefered_value)
                if caps_values.get(field):
                    widget.set_widget_value(caps_values[field])
                self.__add_widget_to_grid(grid, field.capitalize(), widget, y)
                y += 1

                self.caps_widgets[field] = widget

        for y, prop in enumerate(props, start=y):
            # We do not know how to work with GObjects, so blacklist
            # them to avoid noise in the UI
            if (not prop.flags & GObject.ParamFlags.WRITABLE or
                    not prop.flags & GObject.ParamFlags.READABLE or
                    GObject.type_is_a(prop.value_type, GObject.Object)):
                continue

            if is_effect:
                result, prop_value = self.element.get_child_property(prop.name)
                if not result:
                    self.debug(
                        "Could not get value for property: %s", prop.name)
            else:
                if prop.name not in values.keys():
                    # Use the default value.
                    prop_value = self.element.get_property(prop.name)
                else:
                    prop_value = values[prop.name]

            prop_widget = create_property_widget_func(self, prop, prop_value)
            element_name = None
            if isinstance(self.element, Gst.Element):
                element_name = self.element.get_factory().get_name()
            try:
                validation_func = self.INPUT_VALIDATION_FUNCTIONS[(element_name, prop.name)]
                widget = InputValidationWidget(prop_widget, validation_func)
                self.debug("Input validation widget created for (%s, %s)", element_name, prop.name)
            except KeyError:
                widget = prop_widget

            label = Gtk.Label(label=prop.nick)
            label.set_alignment(0.0, 0.5)
            grid.attach(label, 0, y, 1, 1)
            grid.attach(widget, 1, y, 1, 1)
            if hasattr(prop, 'blurb'):
                widget.set_tooltip_text(prop.blurb)

            self.properties[prop] = widget

            if not self.__controllable or isinstance(prop_widget, DefaultWidget):
                continue

            keyframe_button = None
            if not isinstance(prop_widget, (ToggleWidget, ChoiceWidget)):
                res, element, unused_pspec = self.element.lookup_child(prop.name)
                assert res
                binding = GstController.DirectControlBinding.new(
                    element, prop.name,
                    GstController.InterpolationControlSource())
                if binding.pspec:
                    # The prop can be controlled (keyframed).
                    keyframe_button = self.__create_keyframe_toggle_button(prop, widget)
                    grid.attach(keyframe_button, 2, y, 1, 1)

            # The "reset to default" button associated with this property
            if with_reset_button:
                button = self.__create_reset_to_default_button(prop, widget,
                                                               keyframe_button)
                grid.attach(button, 3, y, 1, 1)

        self.element.connect('deep-notify', self._property_changed_cb)
        self.pack_start(grid, expand=False, fill=False, padding=0)
        self.show_all()

    def _make_widget_from_gvalue(self, gvalue, default):
        if isinstance(gvalue, Gst.ValueList):
            choices = []
            for val in gvalue:
                choices.append([val, val])
            widget = ChoiceWidget(choices, default=default[0])
            widget.set_widget_value(default[0])
        else:
            # TODO: implement widgets for other types.
            self.fixme("Unsupported value type: %s", type(gvalue))
            widget = DefaultWidget()

        return widget

    def _property_changed_cb(self, effect, gst_element, pspec):
        if gst_element.get_control_binding(pspec.name):
            self.log("%s controlled, not displaying value", pspec.name)
            return

        widget = self.properties[pspec]
        res, value = self.element.get_child_property(pspec.name)
        assert res
        widget.set_widget_value(value)

    def __create_keyframe_toggle_button(self, prop, widget):
        keyframe_button = Gtk.ToggleButton()
        keyframe_button.props.focus_on_click = False  # Avoid the ugly selection outline
        keyframe_button.set_tooltip_text(_("Show keyframes for this value"))
        keyframe_button.set_relief(Gtk.ReliefStyle.NONE)
        keyframe_button.connect('toggled', self.__keyframes_toggled_cb, prop)
        self.__widgets_by_keyframe_button[keyframe_button] = widget
        prop_binding = self.element.get_control_binding(prop.name)
        self.__bindings_by_keyframe_button[keyframe_button] = prop_binding
        self.__display_controlled(keyframe_button, bool(prop_binding))
        return keyframe_button

    def __create_reset_to_default_button(self, unused_prop, widget,
                                         keyframe_button):
        icon = Gtk.Image()
        icon.set_from_icon_name("edit-clear-all-symbolic", Gtk.IconSize.MENU)
        button = Gtk.Button()
        button.add(icon)
        button.set_tooltip_text(_("Reset to default value"))
        button.set_relief(Gtk.ReliefStyle.NONE)
        button.connect('clicked', self.__reset_to_default_clicked_cb, widget,
                       keyframe_button)
        self.__widgets_by_reset_button[button] = widget
        return button

    def __set_keyframe_active(self, toggle_button, active):
        """Updates the specified button without triggering signals.

        Args:
            toggle_button (Gtk.ToggleButton): The toggle button enabling controlling
                the property by keyframes.
            active (bool): The desired status of the `toggle_button`.
        """
        self.log("Manually resetting the UI state of %s", toggle_button)
        toggle_button.handler_block_by_func(self.__keyframes_toggled_cb)
        toggle_button.set_active(active)
        toggle_button.handler_unblock_by_func(self.__keyframes_toggled_cb)

    def __display_controlled(self, toggle_button, controlled):
        """Displays whether the prop is keyframed."""
        widget = self.__widgets_by_keyframe_button[toggle_button]
        # The displayed value means nothing if the prop is controlled.
        widget.set_sensitive(not controlled)
        # The full disc label indicates the property is controlled (keyframed).
        toggle_button.set_label("◆" if controlled else "◇")

    def __keyframes_toggled_cb(self, keyframe_button, prop):
        self.log("keyframes togglebutton clicked for %s", prop)
        active = keyframe_button.get_active()
        # Now change the state of the *other* togglebuttons.
        for toggle_keyframe_button in self.__widgets_by_keyframe_button:
            if toggle_keyframe_button != keyframe_button:
                # Don't use set_active directly on the buttons; doing so will
                # fire off signals that will toggle the others/confuse the UI
                self.__set_keyframe_active(toggle_keyframe_button, False)
        # The only way to *deactivate* keyframes (not just hide them) is to use
        # the separate reset button.
        self.__display_controlled(keyframe_button, True)

        track_element = self.__get_track_element_of_same_type(self.element)
        if not track_element:
            return

        if active:
            track_element.ui_element.show_keyframes(self.element, prop)
            binding = self.element.get_control_binding(prop.name)
            self.__bindings_by_keyframe_button[keyframe_button] = binding
        else:
            track_element.ui_element.show_default_keyframes()

    def __reset_to_default_clicked_cb(self, unused_button, widget,
                                      keyframe_button=None):
        if keyframe_button:
            # The prop is controllable (keyframmable).
            binding = self.__bindings_by_keyframe_button.get(keyframe_button)
            if binding:
                # The prop has been keyframed
                binding.props.control_source.unset_all()
                if keyframe_button.get_active():
                    track_element = self.__get_track_element_of_same_type(
                        self.element)
                    if track_element:
                        track_element.ui_element.show_default_keyframes()
                self.__set_keyframe_active(keyframe_button, False)
                self.__display_controlled(keyframe_button, False)

        widget.set_widget_to_default()

    def __get_track_element_of_same_type(self, effect):
        track_type = effect.get_track_type()
        for track_element in effect.get_parent().get_children(False):
            if hasattr(track_element, "ui_element") and \
                    track_element.get_track_type() == track_type:
                return track_element
        self.warning("Failed to find track element of type %s", track_type)
        return None

    def get_settings(self, with_default=False):
        """Gets a name/value dict with the properties."""
        values = {}
        for prop, widget in self.properties.items():
            if not prop.flags & GObject.ParamFlags.WRITABLE:
                continue
            value = widget.get_widget_value()
            if value is not None and (value != prop.default_value or with_default):
                values[prop.name] = value
        return values

    def get_caps_values(self):
        values = {}
        for field, widget in self.caps_widgets.items():
            value = widget.get_widget_value()
            if value is not None:
                values[field] = value

        return values

    def make_property_widget(self, prop, value=None):
        """Creates a widget for the specified element property."""
        type_name = GObject.type_name(prop.value_type.fundamental)
        if type_name == "gchararray":
            widget = TextWidget(default=prop.default_value)
        elif type_name in ['guint64', 'gint64', 'guint', 'gint', 'gfloat', 'gulong', 'gdouble']:
            maximum, minimum = None, None
            if hasattr(prop, "minimum"):
                minimum = prop.minimum
            if hasattr(prop, "maximum"):
                maximum = prop.maximum
            widget = NumericWidget(
                default=prop.default_value, upper=maximum, lower=minimum)
        elif type_name == "gboolean":
            widget = ToggleWidget(default=prop.default_value)
        elif type_name == "GEnum":
            choices = []
            for unused_key, val in prop.enum_class.__enum_values__.items():
                choices.append([val.value_name, int(val)])
            widget = ChoiceWidget(choices, default=prop.default_value)
        elif type_name == "GstFraction":
            widget = FractionWidget(
                presets=["0:1"], default=prop.default_value)
        else:
            # TODO: implement widgets for: GBoxed, GFlags
            self.fixme("Unsupported property type: %s", type_name)
            widget = DefaultWidget()

        if value is None:
            value = prop.default_value
        if value is not None:
            widget.set_widget_value(value)

        return widget

    def get_widget_of_prop(self, prop_name):
        for prop in self.properties:
            if prop.name == prop_name:
                return self.properties[prop]
        return None


class GstElementSettingsDialog(Loggable):
    """Dialog window for viewing/modifying properties of a Gst.Element."""

    def __init__(self, elementfactory, properties, parent_window=None,
                 caps=None):
        Loggable.__init__(self)
        self.debug("factory: %s, properties: %s", elementfactory, properties)

        self.factory = elementfactory
        self.element = self.factory.create("elementsettings")
        if not self.element:
            self.warning(
                "Couldn't create element from factory %s", self.factory)
        self.properties = properties
        self.__caps = caps

        self.builder = Gtk.Builder()
        self.builder.add_from_file(
            os.path.join(get_ui_dir(), "elementsettingsdialog.ui"))
        self.builder.connect_signals(self)
        self.ok_btn = self.builder.get_object("okbutton1")

        self.window = self.builder.get_object("dialog1")
        self.elementsettings = GstElementSettingsWidget(self.element, controllable=False)
        self.builder.get_object("viewport1").add(self.elementsettings)

        # set title and frame label
        self.window.set_title(
            _("Properties for %s") % self.factory.get_longname())

        caps_values = {}
        if self.__caps:
            element_name = None
            if isinstance(self.element, Gst.Element):
                element_name = self.element.get_factory().get_name()
            src_caps_fields = GstElementSettingsWidget.CAP_FIELDS_TO_EXPOSE.get(element_name)
            if src_caps_fields:
                for field in src_caps_fields.keys():
                    val = caps[0][field]
                    if val is not None and Gst.value_is_fixed(val):
                        caps_values[field] = val
        self.elementsettings.add_widgets(GstElementSettingsWidget.make_property_widget,
                                         with_reset_button=True,
                                         values=properties,
                                         caps_values=caps_values)
        disable_scroll(self.elementsettings)

        # Try to avoid scrolling, whenever possible.
        screen_height = self.window.get_screen().get_height()
        contents_height = self.elementsettings.get_preferred_size()[0].height
        maximum_contents_height = max(500, 0.7 * screen_height)
        if contents_height < maximum_contents_height:
            # The height of the content is small enough, disable the
            # scrollbars.
            default_height = -1
            scrolledwindow = self.builder.get_object("scrolledwindow1")
            scrolledwindow.set_policy(
                Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
            scrolledwindow.set_shadow_type(Gtk.ShadowType.NONE)
        else:
            # If we need to scroll, set a reasonable height for the window.
            default_height = 600
        self.window.set_default_size(400, default_height)

        if parent_window:
            self.window.set_transient_for(parent_window)
        self.window.show()

    def get_settings(self):
        """Gets the settings of the `element`.

        Returns:
            dict: A property name to value map.
        """
        return self.elementsettings.get_settings()

    def get_caps(self):
        values = self.elementsettings.get_caps_values()
        if self.__caps and values:
            caps = Gst.Caps(self.__caps.to_string())

            for field, value in values.items():
                caps.set_value(field, value)

            return caps
        return None

    def _reset_values_clicked_cb(self, unused_button):
        self.reset_all()

    def reset_all(self):
        for unused_prop, widget in self.elementsettings.properties.items():
            widget.set_widget_to_default()


class ZoomBox(Gtk.Grid, Zoomable):
    """Container holding the widgets for zooming.

    Attributes:
        timeline (TimelineContainer): The timeline container this belongs to.
    """

    def __init__(self, timeline):
        Gtk.Grid.__init__(self)
        Zoomable.__init__(self)

        self._manual_set = False
        self.timeline = timeline

        zoom_fit_btn = Gtk.Button()
        zoom_fit_btn.set_relief(Gtk.ReliefStyle.NONE)
        zoom_fit_btn.set_tooltip_text(_("Zoom Fit"))
        zoom_fit_btn_grid = Gtk.Grid()
        zoom_fit_icon = Gtk.Image.new_from_icon_name(
            "zoom-fit-best-symbolic", Gtk.IconSize.BUTTON)
        zoom_fit_btn_grid.add(zoom_fit_icon)
        zoom_fit_btn_label = Gtk.Label(label=_("Zoom"))
        zoom_fit_btn_grid.add(zoom_fit_btn_label)
        zoom_fit_btn_grid.set_column_spacing(SPACING / 2)
        zoom_fit_btn.add(zoom_fit_btn_grid)
        zoom_fit_btn.connect("clicked", self._zoom_fit_cb)
        self.attach(zoom_fit_btn, 0, 0, 1, 1)

        # zooming slider
        self._zoom_adjustment = Gtk.Adjustment()
        self._zoom_adjustment.props.lower = 0
        self._zoom_adjustment.props.upper = Zoomable.zoom_steps
        zoomslider = Gtk.Scale.new(
            Gtk.Orientation.HORIZONTAL, adjustment=self._zoom_adjustment)
        # Setting _zoom_adjustment's value must be done after we create the
        # zoom slider, otherwise the slider remains at the leftmost position.
        self._zoom_adjustment.set_value(Zoomable.get_current_zoom_level())
        zoomslider.props.draw_value = False
        zoomslider.connect("scroll-event", self._zoom_slider_scroll_cb)
        zoomslider.connect("value-changed", self._zoom_slider_changed_cb)
        zoomslider.connect("query-tooltip", self._zoom_slider_query_tooltip_cb)
        zoomslider.set_has_tooltip(True)
        # At least 100px wide for precision
        zoomslider.set_size_request(100, 0)
        zoomslider.set_hexpand(True)
        self.attach(zoomslider, 1, 0, 1, 1)

        # Empty label so we have some spacing at the right of the zoomslider
        self.attach(Gtk.Label(label=""), 2, 0, 1, 1)

        self.set_hexpand(False)
        self.set_column_spacing(ZOOM_SLIDER_PADDING)
        self.show_all()

    def _zoom_slider_changed_cb(self, adjustment):
        Zoomable.set_zoom_level(adjustment.get_value())
        self.timeline.app.write_action("set-zoom-level",
                                       level=adjustment.get_value(),
                                       optional_action_type=True)

        if not self._manual_set:
            self.timeline.timeline.scroll_to_playhead(delayed=True)

    def _zoom_fit_cb(self, button):
        self.timeline.timeline.set_best_zoom_ratio(allow_zoom_in=True)

    def _zoom_slider_scroll_cb(self, unused, event):
        delta = 0
        if event.direction in [Gdk.ScrollDirection.UP, Gdk.ScrollDirection.RIGHT]:
            delta = 1
        elif event.direction in [Gdk.ScrollDirection.DOWN, Gdk.ScrollDirection.LEFT]:
            delta = -1
        elif event.direction in [Gdk.ScrollDirection.SMOOTH]:
            unused_res, delta_x, delta_y = event.get_scroll_deltas()
            if delta_x:
                delta = math.copysign(1, delta_x)
            elif delta_y:
                delta = math.copysign(1, -delta_y)
        if delta:
            Zoomable.set_zoom_level(Zoomable.get_current_zoom_level() + delta)

    def zoom_changed(self):
        zoom_level = self.get_current_zoom_level()
        if int(self._zoom_adjustment.get_value()) != zoom_level:
            self._manual_set = True
            try:
                self._zoom_adjustment.set_value(zoom_level)
            finally:
                self._manual_set = False

    def _zoom_slider_query_tooltip_cb(self, unused_slider, unused_x, unused_y, unused_keyboard_mode, tooltip):
        # We assume the width of the ruler is exactly the width of the
        # timeline.
        width_px = self.timeline.ruler.get_allocated_width()
        timeline_width_ns = Zoomable.pixel_to_ns(width_px)
        if timeline_width_ns >= Gst.SECOND:
            # Translators: %s represents a duration, for example "10 minutes"
            tip = _("%s displayed") % beautify_length(timeline_width_ns)
        else:
            # Translators: This is a tooltip
            tip = _(
                "%d nanoseconds displayed, because we can") % timeline_width_ns
        tooltip.set_text(tip)
        return True


DROPPER_BITS = (
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377\377\377\377\377\377\377\377"
    b"\377\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377\377\0\0\0\377"
    b"\0\0\0\377\0\0\0\377\377\377\377\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377"
    b"\377\0\0\0\377\0\0\0\377\0\0\0\377\0\0\0\377\0\0\0\377\377\377\377\377"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377"
    b"\377\377\377\377\377\377\377\377\377\377\377\0\0\0\377\0\0\0\377\0\0"
    b"\0\377\0\0\0\377\0\0\0\377\377\377\377\377\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377\377\0\0\0\377\0\0\0\377\0"
    b"\0\0\377\0\0\0\377\0\0\0\377\0\0\0\377\0\0\0\377\0\0\0\377\377\377\377"
    b"\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\377\377\377\377\0\0\0\377\0\0\0\377\0\0\0\377\0\0\0\377\0\0\0\377\0"
    b"\0\0\377\377\377\377\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377\377\377\0\0\0\377\0\0"
    b"\0\377\0\0\0\377\377\377\377\377\377\377\377\377\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377"
    b"\377\377\377\377\377\377\377\377\377\377\0\0\0\377\0\0\0\377\377\377"
    b"\377\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\377\377\377\377\377\377\377\377\377\377\377\377\377"
    b"\0\0\0\377\377\377\377\377\0\0\0\377\377\377\377\377\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377\377"
    b"\377\377\377\377\377\377\377\377\377\0\0\0\377\0\0\0\0\0\0\0\0\377\377"
    b"\377\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\377\377\377\377\377\377\377\377\377\377\377\377\377\0\0\0"
    b"\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377\377\377\377\377\377\377\377"
    b"\377\377\377\0\0\0\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377\377\377"
    b"\377\377\377\377\377\377\377\377\0\0\0\377\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\377\377\377\377\377\377\377\377\377\377\377\377\377\0\0\0\377\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\377\377\377\377\377\377\377\377\0\0"
    b"\0\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\0\0\0\0\0\0\0\377\0\0\0"
    b"\377\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\377\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
    b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0")

DROPPER_WIDTH = 17
DROPPER_HEIGHT = 17
DROPPER_X_HOT = 2
DROPPER_Y_HOT = 16


class ColorPickerButton(Gtk.Button):

    # This ported ColorPickerButton borrows from deprecated GtkColorsel widget

    __gsignals__ = {
        "value-changed": (GObject.SignalFlags.RUN_LAST, None, (),),
    }

    color_r = GObject.Property(type=int, default=0)
    color_g = GObject.Property(type=int, default=0)
    color_b = GObject.Property(type=int, default=0)

    def __init__(self, r=0, g=0, b=0):
        Gtk.Button.__init__(self)
        self.color_r = r
        self.color_g = g
        self.color_b = b
        self.dropper_grab_widget = None
        self.pointer_device = None
        self.is_picking = False
        picker_image = Gtk.Image.new_from_icon_name("gtk-color-picker", Gtk.IconSize.BUTTON)
        self.set_image(picker_image)
        self.connect("clicked", self.clicked_cb)

    def clicked_cb(self, button):
        device = Gtk.get_current_event_device()
        # Allow picking color using mouse only
        if device == Gdk.InputSource.KEYBOARD:
            return
        self.pointer_device = device
        screen = self.get_screen()
        if self.dropper_grab_widget is None:
            grab_widget = Gtk.Window(Gtk.WindowType.POPUP)
            grab_widget.set_screen(screen)
            grab_widget.resize(1, 1)
            grab_widget.move(-100, -100)
            grab_widget.show()
            grab_widget.add_events(Gdk.EventMask.BUTTON_RELEASE_MASK)
            top_level = self.get_toplevel()
            if isinstance(top_level, Gtk.Window):
                if top_level.has_group():
                    top_level.get_group().add_window(grab_widget)
            self.dropper_grab_widget = grab_widget

        window = self.dropper_grab_widget.get_window()
        picker_cursor = self.make_cursor_picker(screen)
        # Connect to button release event else unnecessary events are triggered
        grab_status = self.pointer_device.grab(window, Gdk.GrabOwnership.APPLICATION, False,
                                               Gdk.EventMask.BUTTON_RELEASE_MASK,
                                               picker_cursor, Gdk.CURRENT_TIME)
        if grab_status != Gdk.GrabStatus.SUCCESS:
            return
        # Start tracking the mouse events
        self.dropper_grab_widget.grab_add()
        self.dropper_grab_widget.connect("button-release-event", self.button_release_event_cb)

    def make_cursor_picker(self, screen):
        # Get color-picker cursor if it exists in the current theme else fallback to generating it ourself
        try:
            cursor = Gdk.Cursor.new_from_name(screen.get_display(), "color-picker")
        except TypeError:
            pixbuf = GdkPixbuf.Pixbuf.new_from_data(DROPPER_BITS, GdkPixbuf.Colorspace.RGB, True, 8, DROPPER_WIDTH,
                                                    DROPPER_HEIGHT, DROPPER_WIDTH * 4)
            cursor = Gdk.Cursor.new_from_pixbuf(screen.get_display(), pixbuf, DROPPER_X_HOT, DROPPER_Y_HOT)
        return cursor

    def button_release_event_cb(self, widget, event):
        if event.button != Gdk.BUTTON_PRIMARY:
            return

        self.grab_color_at_pointer(event.get_screen(), event.x_root, event.y_root)
        self.emit("value-changed")
        self.shutdown_eyedropper()

    def grab_color_at_pointer(self, screen, x, y):
        root_window = screen.get_root_window()
        pixbuf = Gdk.pixbuf_get_from_window(root_window, x, y, 1, 1)
        pixels = pixbuf.get_pixels()
        self.color_r = pixels[0]
        self.color_g = pixels[1]
        self.color_b = pixels[2]

    def shutdown_eyedropper(self):
        self.is_picking = False
        self.pointer_device.ungrab(Gdk.CURRENT_TIME)
        self.dropper_grab_widget.grab_remove()
        self.pointer_device = None
        self.dropper_grab_widget = None

    def calculate_argb(self):
        argb = 0
        argb += (1 * 255) * 256 ** 3
        argb += float(self.color_r) * 256 ** 2
        argb += float(self.color_g) * 256 ** 1
        argb += float(self.color_b) * 256 ** 0
        argb = int(argb)
        return argb



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2020, Alexandru Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
from pitivi.utils.misc import round05
from tests import common


class TestMisc(common.TestCase):

    def test_round05(self):
        self.assertEqual(round05(0), 0.5)
        self.assertEqual(round05(0.999), 0.5)

        self.assertEqual(round05(1), 1.5)
        self.assertEqual(round05(1.999), 1.5)

        self.assertEqual(round05(2), 2.5)
        self.assertEqual(round05(2.999), 2.5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Pitivi unit tests."""
import os
import sys
from tempfile import mkdtemp

import gi.overrides


def get_pitivi_dir():
    """Gets the pitivi root directory."""
    tests_dir = os.path.dirname(os.path.abspath(__file__))
    pitivi_dir = os.path.join(tests_dir, os.path.pardir)
    return os.path.abspath(pitivi_dir)


def _prepend_env_paths(**args):
    """Prepends one or more paths to an environment variable."""
    for name, value in args.items():
        if not isinstance(value, list):
            value = [value]

        os.environ[name] = os.pathsep.join(
            value + os.environ.get(name, "").split(
                os.pathsep))


def setup():
    """Sets paths and initializes modules, to be able to run the tests."""
    # Make sure xdg_*_home return temp dirs, to avoid touching
    # the config files of the developer.
    os.environ['XDG_DATA_HOME'] = mkdtemp()
    os.environ['XDG_CONFIG_HOME'] = mkdtemp()
    os.environ['XDG_CACHE_HOME'] = mkdtemp()

    # Make available to configure.py the top level dir.
    pitivi_dir = get_pitivi_dir()
    sys.path.insert(0, pitivi_dir)

    from pitivi import configure

    # Let Gst overrides from our prefix take precedence over any
    # other, making sure they are used.
    local_overrides = os.path.join(configure.LIBDIR,
                                   "python" + sys.version[:3],
                                   "site-packages", "gi", "overrides")
    gi.overrides.__path__.insert(0, local_overrides)

    # Make sure that flatpak gst-python overrides are always used first.
    flatpak_gst_python_path = os.path.join("/app/lib/", "python" + sys.version[:3],
                                           "site-packages", "gi", "overrides")
    if os.path.exists(flatpak_gst_python_path):
        gi.overrides.__path__.insert(0, flatpak_gst_python_path)

    # Make available the compiled C code.
    sys.path.append(configure.BUILDDIR)

    _prepend_env_paths(GST_PRESET_PATH=[os.path.join(pitivi_dir, "data", "videopresets"),
                                        os.path.join(pitivi_dir, "data", "audiopresets")],
                       GST_ENCODING_TARGET_PATH=[os.path.join(pitivi_dir, "data", "encoding-profiles")])
    os.environ.setdefault('PITIVI_TOP_LEVEL_DIR', pitivi_dir)

    # Make sure the modules are initialized correctly.
    from pitivi import check
    check.initialize_modules()
    res = check.check_requirements()

    from pitivi.utils import loggable as log
    log.init('PITIVI_DEBUG')

    return res


if not setup():
    raise ImportError("Could not setup testsuite")



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Millan Castro <m.castrovilarino@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Markers display and management."""
from gettext import gettext as _
from typing import Optional

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from pitivi.utils.loggable import Loggable
from pitivi.utils.pipeline import PipelineError
from pitivi.utils.timeline import Zoomable
from pitivi.utils.ui import SPACING

MARKER_WIDTH = 10


class Marker(Gtk.EventBox, Loggable):
    """Widget representing a marker."""

    def __init__(self, ges_marker):
        Gtk.EventBox.__init__(self)
        Loggable.__init__(self)

        self.add_events(Gdk.EventMask.ENTER_NOTIFY_MASK |
                        Gdk.EventMask.LEAVE_NOTIFY_MASK)

        self.ges_marker = ges_marker
        self.ges_marker.ui = self
        self.position_ns = self.ges_marker.props.position

        self.get_style_context().add_class("Marker")
        self.ges_marker.connect("notify-meta", self._notify_meta_cb)

        self._selected = False

    def do_get_request_mode(self):
        return Gtk.SizeRequestMode.CONSTANT_SIZE

    def do_get_preferred_height(self):
        return MARKER_WIDTH, MARKER_WIDTH

    def do_get_preferred_width(self):
        return MARKER_WIDTH, MARKER_WIDTH

    def do_enter_notify_event(self, unused_event):
        self.set_state_flags(Gtk.StateFlags.PRELIGHT, clear=False)

    def do_leave_notify_event(self, unused_event):
        self.unset_state_flags(Gtk.StateFlags.PRELIGHT)

    def _notify_meta_cb(self, unused_ges_marker, item, value):
        self.set_tooltip_text(self.comment)

    @property
    def position(self):
        """Returns the position of the marker, in nanoseconds."""
        return self.ges_marker.props.position

    @property
    def comment(self):
        """Returns a comment from ges_marker."""
        return self.ges_marker.get_string("comment")

    @comment.setter
    def comment(self, text):
        if text == self.comment:
            return
        self.ges_marker.set_string("comment", text)

    @property
    def selected(self):
        """Returns whether the marker is selected."""
        return self._selected

    @selected.setter
    def selected(self, selected):
        self._selected = selected
        if self._selected:
            self.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)
        else:
            self.unset_state_flags(Gtk.StateFlags.SELECTED)


class MarkersBox(Gtk.EventBox, Zoomable, Loggable):
    """Container for displaying and managing markers."""

    def __init__(self, app, hadj=None):
        Gtk.EventBox.__init__(self)
        Zoomable.__init__(self)
        Loggable.__init__(self)

        self.layout = Gtk.Layout()
        self.add(self.layout)
        self.get_style_context().add_class("MarkersBox")

        self.app = app

        if hadj:
            hadj.connect("value-changed", self._hadj_value_changed_cb)
        self.props.hexpand = True
        self.props.valign = Gtk.Align.START

        self.offset = 0
        self.props.height_request = MARKER_WIDTH

        self.__markers_container = None
        self.marker_moving = None
        self.marker_new = None

        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK |
                        Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK)

        self._create_actions()

    def _create_actions(self):
        self.action_group = Gio.SimpleActionGroup()
        self.insert_action_group("markers", self.action_group)
        self.app.shortcuts.register_group("markers", _("Markers"), position=70)

        self.add_marker_action = Gio.SimpleAction.new("marker-add", GLib.VariantType("mx"))
        self.add_marker_action.connect("activate", self._add_marker_cb)
        self.action_group.add_action(self.add_marker_action)
        self.app.shortcuts.add("markers.marker-add(@mx nothing)", ["<Primary><Shift>m"],
                               self.add_marker_action,
                               _("Add a marker"))

        self.seek_backward_marker_action = Gio.SimpleAction.new("seek-backward-marker", None)
        self.seek_backward_marker_action.connect("activate", self._seek_backward_marker_cb)
        self.action_group.add_action(self.seek_backward_marker_action)
        self.app.shortcuts.add("markers.seek-backward-marker", ["<Alt>Left"],
                               self.seek_backward_marker_action,
                               _("Seek to the first marker before the playhead"))

        self.seek_forward_marker_action = Gio.SimpleAction.new("seek-forward-marker", None)
        self.seek_forward_marker_action.connect("activate", self._seek_forward_marker_cb)
        self.action_group.add_action(self.seek_forward_marker_action)
        self.app.shortcuts.add("markers.seek-forward-marker", ["<Alt>Right"],
                               self.seek_forward_marker_action,
                               _("Seek to the first marker after the playhead"))

    def _seek_backward_marker_cb(self, action, param):
        current_position = self.app.project_manager.current_project.pipeline.get_position(fails=False)
        position = self.first_marker(before=current_position)
        if position is None:
            return

        self.app.project_manager.current_project.pipeline.simple_seek(position)
        self.app.gui.editor.timeline_ui.timeline.scroll_to_playhead(align=Gtk.Align.CENTER, when_not_in_view=True)

    def _seek_forward_marker_cb(self, action, param):
        current_position = self.app.project_manager.current_project.pipeline.get_position(fails=False)
        position = self.first_marker(after=current_position)
        if position is None:
            return

        self.app.project_manager.current_project.pipeline.simple_seek(position)
        self.app.gui.editor.timeline_ui.timeline.scroll_to_playhead(align=Gtk.Align.CENTER, when_not_in_view=True)

    def first_marker(self, before: Optional[int] = None, after: Optional[int] = None) -> Optional[int]:
        assert (after is not None) != (before is not None)

        if after is not None:
            start = after + 1
            end = self.app.project_manager.current_project.ges_timeline.props.duration
        else:
            start = 0
            end = before

        if start >= end:
            return None

        markers_positions = list([ges_marker.props.position
                                  for ges_marker in self.__markers_container.get_markers()
                                  if start <= ges_marker.props.position < end])
        if not markers_positions:
            return None

        if after is not None:
            return min(markers_positions)
        else:
            return max(markers_positions)

    def _add_marker_cb(self, action, param):
        maybe = param.get_maybe()
        if maybe:
            position = maybe.get_int64()
        else:
            try:
                position = self.app.project_manager.current_project.pipeline.get_position(fails=False)
            except PipelineError:
                self.warning("Could not get pipeline position")
                return

        with self.app.action_log.started("Added marker", toplevel=True):
            self.__markers_container.add(position)

    @property
    def markers_container(self):
        """Gets the GESMarkerContainer."""
        return self.__markers_container

    @markers_container.setter
    def markers_container(self, ges_markers_container):
        if self.__markers_container:
            for marker in self.layout.get_children():
                self.layout.remove(marker)
            self.__markers_container.disconnect_by_func(self._marker_added_cb)

        self.__markers_container = ges_markers_container
        if self.__markers_container:
            self.__create_marker_widgets()
            self.__markers_container.connect("marker-added", self._marker_added_cb)
            self.__markers_container.connect("marker-removed", self._marker_removed_cb)
            self.__markers_container.connect("marker-moved", self._marker_moved_cb)

    def __create_marker_widgets(self):
        markers = self.__markers_container.get_markers()

        for ges_marker in markers:
            position = ges_marker.props.position
            self._add_marker(position, ges_marker)

    def _hadj_value_changed_cb(self, hadj):
        """Handles the adjustment value change."""
        self.offset = hadj.get_value()
        self._update_position()

    def zoom_changed(self):
        self._update_position()

    def _update_position(self):
        for marker in self.layout.get_children():
            position = self.ns_to_pixel(marker.position) - self.offset - MARKER_WIDTH / 2
            self.layout.move(marker, position, 0)

    def do_button_press_event(self, event):
        event_widget = Gtk.get_event_widget(event)
        button = event.button
        if button == Gdk.BUTTON_PRIMARY:
            if isinstance(event_widget, Marker):
                if event.type == Gdk.EventType.BUTTON_PRESS:
                    self.marker_moving = event_widget
                    self.marker_moving.selected = True
                    self.app.action_log.begin("Move marker", toplevel=True)

                elif event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
                    self.marker_moving = None
                    self.app.action_log.rollback()
                    marker_popover = MarkerPopover(self.app, event_widget)
                    marker_popover.popup()

            else:
                position = self.pixel_to_ns(event.x + self.offset)
                param = GLib.Variant.new_maybe(GLib.VariantType("x"), GLib.Variant.new_int64(position))
                self.add_marker_action.activate(param)
                self.marker_new.selected = True

    def do_button_release_event(self, event):
        button = event.button
        event_widget = Gtk.get_event_widget(event)
        if button == Gdk.BUTTON_PRIMARY:
            if self.marker_moving:
                self.marker_moving.selected = False
                self.marker_moving = None
                self.app.action_log.commit("Move marker")
            elif self.marker_new:
                self.marker_new.selected = False
                self.marker_new = None

        elif button == Gdk.BUTTON_SECONDARY and isinstance(event_widget, Marker):
            with self.app.action_log.started("Removed marker", toplevel=True):
                self.__markers_container.remove(event_widget.ges_marker)

    def do_motion_notify_event(self, event):
        event_widget = Gtk.get_event_widget(event)
        if event_widget is self.marker_moving:
            event_x, unused_y = event_widget.translate_coordinates(self, event.x, event.y)
            event_x = max(0, event_x)
            position_ns = self.pixel_to_ns(event_x + self.offset)
            self.__markers_container.move(self.marker_moving.ges_marker, position_ns)

    def _marker_added_cb(self, unused_markers, position, ges_marker):
        self._add_marker(position, ges_marker)

    def _add_marker(self, position, ges_marker):
        marker = Marker(ges_marker)
        x = self.ns_to_pixel(position) - self.offset - MARKER_WIDTH / 2
        self.layout.put(marker, x, 0)
        marker.show()
        self.marker_new = marker

    def _marker_removed_cb(self, unused_markers, ges_marker):
        self._remove_marker(ges_marker)

    def _remove_marker(self, ges_marker):
        if not ges_marker.ui:
            return

        self.layout.remove(ges_marker.ui)
        ges_marker.ui = None

    def _marker_moved_cb(
            self, unused_markers, unused_prev_position, position, ges_marker):
        self._move_marker(position, ges_marker)

    def _move_marker(self, position, ges_marker):
        x = self.ns_to_pixel(position) - self.offset - MARKER_WIDTH / 2
        self.layout.move(ges_marker.ui, x, 0)


class MarkerPopover(Gtk.Popover):
    """A popover to edit a marker's metadata."""

    def __init__(self, app, marker):
        Gtk.Popover.__init__(self)

        self.app = app

        self.text_view = Gtk.TextView()
        self.text_view.set_size_request(100, -1)

        self.marker = marker

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.props.margin = SPACING
        vbox.pack_start(self.text_view, False, True, 0)
        self.add(vbox)

        text = self.marker.comment
        if text:
            text_buffer = self.text_view.get_buffer()
            text_buffer.set_text(text)

        self.set_position(Gtk.PositionType.TOP)
        self.set_relative_to(self.marker)
        self.show_all()

    def do_closed(self):
        buffer = self.text_view.get_buffer()
        if buffer.props.text != self.marker.comment:
            with self.app.action_log.started("marker comment", toplevel=True):
                self.marker.comment = buffer.props.text
        self.marker.selected = False



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (C) 2004,2005,2006,2007 Fluendo, S.L. (www.fluendo.com).
# All rights reserved.
#
# This file may be distributed and/or modified under the terms of
# the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# This file is distributed without any warranty; without even the implied
# warranty of merchantability or fitness for a particular purpose.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import unittest

from pitivi.utils import loggable as log


class LogTester(log.Loggable):
    log_category = 'testlog'


class LogFunctionTester(log.Loggable):

    def log_function(self, fmt, *args):
        return (("override " + fmt), ) + args[1:]


class TestWithHandler(unittest.TestCase):

    def setUp(self):
        self.level = None
        self.object = None
        self.category = None
        self.file = None
        self.line = None
        self.message = None

        # we want to remove the default handler so it doesn't show up stuff
        log.reset()

    def handler(self, level, obj, category, file, line, message):
        self.level = level
        self.object = obj
        self.category = category
        self.file = file
        self.line = line
        self.message = message.split(' ', 1)[1]


class TestLog(TestWithHandler):

    def setUp(self):
        TestWithHandler.setUp(self)
        self.tester = LogTester()

    # just test for parsing semi- or non-valid FLU_DEBUG variables

    def test_set_debug(self):
        log.set_debug(":5")
        log.set_debug("*")
        log.set_debug("5")

    # test for adding a log handler

    def test_limit_invisible(self):
        log.set_debug("testlog:%d" % log.INFO)
        log.add_limited_log_handler(self.handler)

        # log 2 we shouldn't get
        self.tester.log("not visible")
        self.assertFalse(self.category)
        self.assertFalse(self.level)
        self.assertFalse(self.message)

        self.tester.debug("not visible")
        self.assertFalse(self.category)
        self.assertFalse(self.level)
        self.assertFalse(self.message)

    def test_limited_visible(self):
        log.set_debug("testlog:%d" % log.INFO)
        log.add_limited_log_handler(self.handler)

        # log 3 we should get
        self.tester.info("visible")
        self.assertEqual(self.category, 'testlog')
        self.assertEqual(self.level, log.INFO)
        self.assertEqual(self.message, 'visible')

        self.tester.warning("also visible")
        self.assertEqual(self.category, 'testlog')
        self.assertEqual(self.level, log.WARN)
        self.assertEqual(self.message, 'also visible')

    def test_format_strings(self):
        log.set_debug("testlog:%d" % log.INFO)
        log.add_limited_log_handler(self.handler)

        self.tester.info("%d %s", 42, 'the answer')
        self.assertEqual(self.category, 'testlog')
        self.assertEqual(self.level, log.INFO)
        self.assertEqual(self.message, '42 the answer')

    def test_limited_error(self):
        log.set_debug("testlog:%d" % log.ERROR)
        log.add_limited_log_handler(self.handler)

        self.tester.error("error")
        self.assertEqual(self.category, 'testlog')
        self.assertEqual(self.level, log.ERROR)
        self.assertEqual(self.message, 'error')

    def test_log_handler_limited_levels(self):
        log.set_debug("testlog:%d" % log.INFO)
        log.add_limited_log_handler(self.handler)

        # now try debug and log again too
        log.set_debug("testlog:%d" % log.LOG)

        self.tester.debug("debug")
        self.assertEqual(self.category, 'testlog')
        self.assertEqual(self.level, log.DEBUG)
        self.assertEqual(self.message, 'debug')

        self.tester.log("log")
        self.assertEqual(self.category, 'testlog')
        self.assertEqual(self.level, log.LOG)
        self.assertEqual(self.message, 'log')

    # test that we get all log messages

    def test_log_handler(self):
        log.set_debug("testlog:%d" % log.INFO)
        log.add_log_handler(self.handler)

        self.tester.log("visible")
        self.assertEqual(self.message, 'visible')

        self.tester.warning("also visible")
        self.assertEqual(self.message, 'also visible')


class TestOwnLogHandler(TestWithHandler):

    def setUp(self):
        TestWithHandler.setUp(self)
        self.tester = LogFunctionTester()

    def test_own_log_handler_limited(self):
        """Checks our own log handler correctly mangles the message."""
        log.set_debug("testlog:%d" % log.INFO)
        log.add_log_handler(self.handler)

        self.tester.log("visible")
        self.assertEqual(self.message, 'override visible')

    def test_log_handler_assertion(self):
        self.assertRaises(TypeError, log.add_limited_log_handler, None)


class TestGetExceptionMessage(unittest.TestCase):

    def func3(self):
        self.func2()

    def func2(self):
        self.func1()

    def func1(self):
        raise TypeError("I am in func1")

    def test_level2(self):
        try:
            self.func2()
            self.fail("Should not get to this point")
        except TypeError as e:
            self.verify_exception(e)

    def test_level3(self):
        try:
            self.func3()
            self.fail("Should not get to this point")
        except TypeError as e:
            self.verify_exception(e)

    def verify_exception(self, e):
        message = log.get_exception_message(e)
        self.assertTrue("func1()" in message, message)
        self.assertTrue("test_utils_loggable.py" in message, message)
        self.assertTrue("TypeError" in message, message)


class TestLogSettings(unittest.TestCase):

    def test_set(self):
        old = log.get_log_settings()
        log.set_debug('*:5')
        self.assertNotEqual(old, log.get_log_settings())

        log.set_log_settings(old)
        self.assertEqual(old, log.get_log_settings())


class TestLogNames(unittest.TestCase):

    def test_get_level_code(self):
        self.assertEqual(1, log.get_level_int('ERROR'))
        self.assertEqual(2, log.get_level_int('WARN'))
        self.assertEqual(3, log.get_level_int('FIXME'))
        self.assertEqual(4, log.get_level_int('INFO'))
        self.assertEqual(5, log.get_level_int('DEBUG'))
        self.assertEqual(6, log.get_level_int('LOG'))

    def test_get_level_name(self):
        self.assertEqual('ERROR', log.get_level_name(1))
        self.assertEqual('WARN', log.get_level_name(2))
        self.assertEqual('FIXME', log.get_level_name(3))
        self.assertEqual('INFO', log.get_level_name(4))
        self.assertEqual('DEBUG', log.get_level_name(5))
        self.assertEqual('LOG', log.get_level_name(6))



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2013, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
from gi.repository import Gdk
from gi.repository import Gst

from pitivi.utils.widgets import ChoiceWidget
from pitivi.utils.widgets import ColorWidget
from pitivi.utils.widgets import FontWidget
from pitivi.utils.widgets import FractionWidget
from pitivi.utils.widgets import GstElementSettingsDialog
from pitivi.utils.widgets import NumericWidget
from pitivi.utils.widgets import PathWidget
from pitivi.utils.widgets import TextWidget
from pitivi.utils.widgets import ToggleWidget
from tests import common


class TestWidgets(common.TestCase):

    def test_construction(self):
        widgets = (
            (PathWidget, "file:///home/", ()),
            (TextWidget, "banana", ()),
            (NumericWidget, 42, (100, 1)),
            (ToggleWidget, True, ()),
            (ChoiceWidget, "banana", ((
                ("banana", "banana"),
                ("apple", "apple"),
                ("pear", "pear")),)),
            (ColorWidget, Gdk.RGBA(0.5, 0.5, 0.3, 0.8), ()),
            (FontWidget, "Sans 9", ()))

        for widget_class, default, args in widgets:
            widget = widget_class(*args, default=default)
            self.assertEqual(default, widget.get_widget_default())
            widget.set_widget_to_default()
            self.assertEqual(default, widget.get_widget_value())
            widget.set_widget_value(default)
            self.assertEqual(default, widget.get_widget_value())

    def test_validation(self):
        widget = TextWidget("^([a-zA-Z]+\\s*)+$")
        bad_value = "1"
        self.assertNotEqual(bad_value, widget.get_widget_value())

        widget = TextWidget("^\\d+$", ("12", "14"))
        bad_value = "non-digits"
        self.assertNotEqual(bad_value, widget.get_widget_value())


class TestFractionWidget(common.TestCase):

    def test_widget_text(self):
        widget = FractionWidget()
        widget.set_widget_value(Gst.Fraction(1000000, 1))
        self.assertEqual(widget.text.get_text(), "1000000:1")
        widget.set_widget_value(Gst.Fraction(7504120000000001, 4503600000000002))
        self.assertEqual(widget.text.get_text(), "7504120000000001:4503600000000002")

    def test_widget_text_magic__m(self):
        widget = FractionWidget()
        widget.set_widget_value(Gst.Fraction(1000000000, 1001))
        self.assertEqual(widget.text.get_text(), "1000000M")


class TestGstElementSettingsDialog(common.TestCase):

    def test_reusing_properties(self):
        """Checks passing values to be used on element to be configured works."""
        values = {"datarate": 12}
        dialog = GstElementSettingsDialog(Gst.ElementFactory.find("identity"),
                                          values)
        widgets = {prop.name: widget
                   for prop, widget in dialog.elementsettings.properties.items()}
        self.assertEqual(widgets["datarate"].get_widget_value(), 12)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2010, Thibault Saunier <tsaunier@gnome.org>
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Effects categorization and management.

There are different types of effects available:
  _ Simple Audio/Video Effects
     GStreamer elements that only apply to audio OR video
     Only take the elements who have a straightforward meaning/action
  _ Expanded Audio/Video Effects
     These are the Gstreamer elements that don't have a easy meaning/action or
     that are too cumbersome to use as such
  _ Complex Audio/Video Effects
"""
import os
import re
import subprocess
import sys
import threading
from gettext import gettext as _

import cairo
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GES
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import Gtk

from pitivi.configure import get_pixmap_dir
from pitivi.configure import get_ui_dir
from pitivi.settings import GlobalSettings
from pitivi.utils.loggable import Loggable
from pitivi.utils.ui import disable_scroll
from pitivi.utils.ui import EFFECT_TARGET_ENTRY
from pitivi.utils.ui import PADDING
from pitivi.utils.ui import SPACING
from pitivi.utils.widgets import FractionWidget
from pitivi.utils.widgets import GstElementSettingsWidget

(VIDEO_EFFECT, AUDIO_EFFECT) = list(range(1, 3))

ALLOWED_ONLY_ONCE_EFFECTS = ['videoflip']

EFFECTS_CATEGORIES = (
    (_("Colors"), (
        # Mostly "serious" stuff that relates to correction/adjustments
        # Fancier stuff goes into the "fancy" category
        'burn', 'chromahold', 'cogcolorspace', 'coloreffects', 'dodge',
        'exclusion', 'frei0r-filter-3-point-color-balance',
        'frei0r-filter-brightness', 'frei0r-filter-bw0r',
        'frei0r-filter-color-distance', 'frei0r-filter-coloradj-rgb',
        'frei0r-filter-contrast0r', 'frei0r-filter-curves',
        'frei0r-filter-equaliz0r', 'frei0r-filter-gamma', 'frei0r-filter-glow',
        'frei0r-filter-hueshift0r', 'frei0r-filter-invert0r',
        'frei0r-filter-levels', 'frei0r-filter-primaries',
        'frei0r-filter-saturat0r', 'frei0r-filter-sop-sat',
        'frei0r-filter-threelay0r', 'frei0r-filter-threshold0r',
        'frei0r-filter-tint0r', 'frei0r-filter-twolay0r',
        'frei0r-filter-white-balance', 'gamma', 'videobalance', 'videomedian',
    )),
    (_("Compositing"), (
        'alpha', 'alphacolor', 'frei0r-filter-alpha0ps',
        'frei0r-filter-alphagrad', 'frei0r-filter-alphaspot',
        'frei0r-filter-bluescreen0r', 'frei0r-filter-mask0mate',
        'frei0r-filter-select0r', 'frei0r-filter-transparency',
        'gdkpixbufoverlay',
    )),
    (_("Noise & blur"), (
        'diffuse', 'dilate', 'frei0r-filter-edgeglow', 'frei0r-filter-facebl0r',
        'frei0r-filter-hqdn3d', 'frei0r-filter-sharpness',
        'frei0r-filter-squareblur', 'gaussianblur', 'marble', 'smooth',
    )),
    (_("Analysis"), (
        'frei0r-filter-b', 'frei0r-filter-g', 'frei0r-filter-luminance',
        'frei0r-filter-opencvfacedetect', 'frei0r-filter-pr0be',
        'frei0r-filter-pr0file', 'frei0r-filter-r', 'frei0r-filter-rgb-parade',
        'frei0r-filter-vectorscope', 'navigationtest', 'revtv', 'videoanalyse',
        'videodetect', 'videomark',
    )),
    (_("Geometry"), (
        'aspectratiocrop', 'bulge', 'circle', 'cogdownsample', 'cogscale',
        'fisheye', 'frei0r-filter-3dflippo', 'frei0r-filter-c0rners',
        'frei0r-filter-defish0r', 'frei0r-filter-flippo',
        'frei0r-filter-k-means-clustering', 'frei0r-filter-lens-correction',
        'frei0r-filter-letterb0xed', 'frei0r-filter-perspective',
        'frei0r-filter-pixeliz0r', 'frei0r-filter-scale0tilt', 'gdkpixbufscale',
        'kaleidoscope', 'mirror', 'pinch', 'rotate', 'sphere', 'square',
        'stretch', 'twirl', 'videobox', 'videocrop', 'videoflip', 'videoscale',
        'waterriple',
    )),
    (_("Fancy"), (
        'agingtv', 'bulge', 'chromium', 'circle', 'dicetv', 'edgetv', 'fisheye',
        'frei0r-filter-cartoon', 'frei0r-filter-delay0r',
        'frei0r-filter-distort0r', 'frei0r-filter-k-means-clustering',
        'frei0r-filter-light-graffiti', 'frei0r-filter-nosync0r',
        'frei0r-filter-sobel', 'frei0r-filter-tehroxx0r',
        'frei0r-filter-vertigo', 'frei0r-filter-water', 'glfiltersobel',
        'kaleidoscope', 'mirror', 'optv', 'pinch', 'quarktv', 'radioactv',
        'rippletv', 'shagadelictv', 'solarize', 'sphere', 'square', 'streaktv',
        'stretch', 'tunnel', 'twirl', 'vertigotv', 'warptv', 'waterripple',
    )),
    (_("Time"), (
        'frei0r-filter-baltan', 'frei0r-filter-delay0r',
        'frei0r-filter-nervous', 'videorate',
    )),
    (_("Audio"), (
        "pitch", "freeverb", "removesilence", "festival", "speed",
        "audiorate", "volume", "equalizer-nbands", "equalizer-3bands",
        "equalizer-10bands", "rglimiter", "rgvolume", "audiopanorama",
        "audioinvert", "audiokaraoke", "audioamplify", "audiodynamic",
        "audiocheblimit", "audiochebband", "audioiirfilter", "audiowsinclimit",
        "audiowsincband", "audiofirfilter", "audioecho", "scaletempo", "stereo",
        'audioamplify', 'audiochebband', 'audiocheblimit', 'audiodynamic',
        'audioecho', 'audiofirfilter', 'audioiirfilter', 'audioinvert',
        'audiokaraoke', 'audiopanorama', 'audiorate', 'audiowsincband',
        'audiowsinclimit', 'equalizer-10bands', 'equalizer-3bands',
        'equalizer-nbands', 'festival', 'freeverb', 'pitch', 'removesilence',
        'rglimiter', 'rgvolume', 'scaletempo', 'speed', 'stereo', 'volume',
    )),
)

BLACKLISTED_EFFECTS = [
    "alphacolor",
    "cogcolorspace",
    "coglogoinsert",
    "colorconvert",
    "festival",
    "navigationtest",
    "videoanalyse",
    # We prefer to use videocrop, see https://gitlab.gnome.org/GNOME/pitivi/issues/2150
    "videobox",
    "videodetect",
    "volume",
]

HIDDEN_EFFECTS = [
    # Overlaying an image onto a video stream can already be done.
    "gdkpixbufoverlay"]

GlobalSettings.add_config_section('effect-library')
GlobalSettings.add_config_option('favourite_effects',
                                 section='effect-library',
                                 key='favourite-effects',
                                 default=[])

ICON_WIDTH = 80
ICON_HEIGHT = 45


class EffectInfo:
    """Info for displaying and using an effect.

    Attributes:
        effect_name (str): The bin_description identifying the effect.
    """

    def __init__(self, effect_name, media_type, categories,
                 human_name, description):
        object.__init__(self)
        self.effect_name = effect_name
        self.media_type = media_type
        self.categories = categories
        self.description = description
        self.human_name = human_name

    @property
    def icon(self):
        pixdir = os.path.join(get_pixmap_dir(), "effects")
        try:
            # We can afford to scale the images here, the impact is negligible
            icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
                os.path.join(pixdir, self.effect_name + ".png"),
                ICON_WIDTH, ICON_WIDTH)
        except GLib.Error:
            icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
                os.path.join(pixdir, "defaultthumbnail.svg"), ICON_WIDTH, ICON_HEIGHT)

        return icon

    @property
    def bin_description(self):
        """Gets the bin description which defines this effect."""
        if self.effect_name.startswith("gl"):
            return "glupload ! %s ! gldownload" % self.effect_name
        else:
            return self.effect_name

    @staticmethod
    def name_from_bin_description(bin_description):
        """Gets the name of the effect defined by the `bin_description`."""
        if bin_description.startswith("glupload"):
            return bin_description.split("!")[1].strip()
        else:
            return bin_description

    def good_for_track_element(self, track_element):
        """Checks the effect is compatible with the specified track element.

        Args:
            track_element (GES.TrackElement): The track element to check against.

        Returns:
            bool: Whether it makes sense to apply the effect to the track element.
        """
        track_type = track_element.get_track_type()
        if track_type == GES.TrackType.AUDIO:
            return self.media_type == AUDIO_EFFECT
        elif track_type == GES.TrackType.VIDEO:
            return self.media_type == VIDEO_EFFECT
        else:
            return False


class EffectsManager(Loggable):
    """Keeps info about effects and their categories.

    Attributes:
        video_effects (List[Gst.ElementFactory]): The available video effects.
        audio_effects (List[Gst.ElementFactory]): The available audio effects.
    """

    def __init__(self):
        Loggable.__init__(self)
        self.video_effects = []
        self.audio_effects = []
        self.gl_effects = []
        self._effects = {}

        useless_words = ["Video", "Audio", "audio", "effect",
                         _("Video"), _("Audio"), _("Audio").lower(), _("effect")]
        useless_re = re.compile(" |".join(useless_words))

        registry = Gst.Registry.get()
        factories = registry.get_feature_list(Gst.ElementFactory)
        longnames = set()
        duplicate_longnames = set()
        for factory in factories:
            longname = factory.get_longname()
            if longname in longnames:
                duplicate_longnames.add(longname)
            else:
                longnames.add(longname)
        for factory in factories:
            klass = factory.get_klass()
            name = factory.get_name()
            if ("Effect" not in klass or
                    any(black in name for black in BLACKLISTED_EFFECTS)):
                continue

            media_type = None
            if "Audio" in klass:
                self.audio_effects.append(factory)
                media_type = AUDIO_EFFECT
            elif "Video" in klass:
                self.video_effects.append(factory)
                media_type = VIDEO_EFFECT
            if not media_type:
                HIDDEN_EFFECTS.append(name)
                continue

            longname = factory.get_longname()
            if longname in duplicate_longnames:
                # Workaround https://bugzilla.gnome.org/show_bug.cgi?id=760566
                # Add name which identifies the element and is unique.
                longname = "%s %s" % (longname, name)
            human_name = useless_re.sub("", longname).title()
            effect = EffectInfo(name,
                                media_type,
                                categories=self._get_effect_categories(name),
                                human_name=human_name,
                                description=factory.get_description())
            self._effects[name] = effect

        gl_element_factories = registry.get_feature_list_by_plugin("opengl")
        self.gl_effects = [element_factory.get_name()
                           for element_factory in gl_element_factories]
        if self.gl_effects:
            if "gleffects" in os.environ.get("PITIVI_UNSTABLE_FEATURES", ""):
                thread = threading.Thread(target=self._check_gleffects)
                thread.start()
            else:
                HIDDEN_EFFECTS.extend(self.gl_effects)

    def _check_gleffects(self):
        check_pipeline_path = os.path.join(os.path.dirname(__file__), "utils", "check_pipeline.py")
        try:
            res = subprocess.check_output([sys.executable,
                                           check_pipeline_path,
                                           "videotestsrc ! glupload ! gleffects ! fakesink"])
            self.debug(res)
        except subprocess.CalledProcessError as e:
            self.error("Can not use GL effects: %s", e)
            HIDDEN_EFFECTS.extend(self.gl_effects)

    def get_info(self, bin_description):
        """Gets the info for an effect which can be applied.

        Args:
            bin_description (str): The bin_description defining the effect.

        Returns:
            EffectInfo: The info corresponding to the name, or None.
        """
        name = EffectInfo.name_from_bin_description(bin_description)
        return self._effects.get(name)

    def _get_effect_categories(self, effect_name):
        """Gets the categories to which the specified effect belongs.

        Args:
            effect_name (str): The bin_description identifying the effect.

        Returns:
            List[str]: The categories which contain the effect.
        """
        categories = []
        for category_name, effects in EFFECTS_CATEGORIES:
            if effect_name in effects:
                categories.append(category_name)
        if not categories:
            categories.append(_("Uncategorized"))
        return categories

    @property
    def categories(self):
        """Gets the name of all effect categories."""
        return EffectsManager._get_categories_names(EFFECTS_CATEGORIES)

    @staticmethod
    def _get_categories_names(categories):
        ret = [category_name for category_name, unused_effects in categories]
        ret.sort()
        if categories:
            # Add Uncategorized only if there are other categories defined.
            ret.append(_("Uncategorized"))
        return ret


# ----------------------- UI classes to manage effects -------------------------#


class EffectListWidget(Gtk.Box, Loggable):
    """Widget for listing effects."""

    def __init__(self, instance):
        Gtk.Box.__init__(self)
        Loggable.__init__(self)

        self.app = instance

        self._drag_icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
            os.path.join(get_pixmap_dir(), "effects", "defaultthumbnail.svg"),
            ICON_WIDTH, ICON_HEIGHT)
        self._star_icon_regular = GdkPixbuf.Pixbuf.new_from_file_at_size(
            os.path.join(get_pixmap_dir(), "star-regular.svg"), 15, 15)
        self._star_icon_solid = GdkPixbuf.Pixbuf.new_from_file_at_size(
            os.path.join(get_pixmap_dir(), "star-solid.svg"), 15, 15)

        self.set_orientation(Gtk.Orientation.VERTICAL)
        builder = Gtk.Builder()
        builder.add_from_file(os.path.join(get_ui_dir(), "effectslibrary.ui"))
        builder.connect_signals(self)
        toolbar = builder.get_object("effectslibrary_toolbar")
        toolbar.get_style_context().add_class(Gtk.STYLE_CLASS_INLINE_TOOLBAR)
        self.search_entry = builder.get_object("search_entry")
        self.fav_view_toggle = builder.get_object("favourites_toggle")
        self.fav_view_toggle.set_image(Gtk.Image.new_from_pixbuf(self._star_icon_solid))

        self.main_view = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.category_view = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # Used for showing search results and favourites
        self.search_view = Gtk.ListBox(activate_on_single_click=False)
        self.search_view.connect("row-activated", self.effects_listbox_row_activated_cb)

        placeholder_text = Gtk.Label(_("No effects"))
        placeholder_text.props.visible = True
        self.search_view.set_placeholder(placeholder_text)

        self.main_view.pack_start(self.category_view, True, True, 0)
        self.main_view.pack_start(self.search_view, True, True, 0)

        scrollwin = Gtk.ScrolledWindow()
        scrollwin.props.hscrollbar_policy = Gtk.PolicyType.NEVER
        scrollwin.props.vscrollbar_policy = Gtk.PolicyType.AUTOMATIC
        scrollwin.add(self.main_view)

        self.pack_start(toolbar, False, False, 0)
        self.pack_start(scrollwin, True, True, 0)

        # Delay the loading of the available effects so the application
        # starts faster.
        GLib.idle_add(self._load_available_effects_cb)

        scrollwin.show_all()
        toolbar.show_all()
        self.search_view.hide()

    def _load_available_effects_cb(self):
        self._set_up_category_view()
        self.add_effects_to_listbox(self.search_view)

    def _set_up_category_view(self):
        """Adds expanders and effects to the category view."""
        # Add category expanders
        for category in self.app.effects.categories:
            widget = self._create_category_widget(category)
            self.category_view.add(widget)

        # Add effects to category expanders
        for expander in self.category_view.get_children():
            listbox = expander.get_child()
            category_name = expander.get_label()

            self.add_effects_to_listbox(listbox, category_name)

        self.category_view.show_all()

    def add_effects_to_listbox(self, listbox, category=None, only_text=False):
        """Adds effect rows to the given listbox."""
        effects = self.app.effects.video_effects + self.app.effects.audio_effects
        for effect in effects:
            name = effect.get_name()

            if name in HIDDEN_EFFECTS:
                continue

            effect_info = self.app.effects.get_info(name)

            if not category or category in effect_info.categories:
                widget = self._create_effect_widget(name, only_text)
                listbox.add(widget)

    def _create_category_widget(self, category):
        """Creates an expander for the given category."""
        expander = Gtk.Expander(label=category, margin=SPACING)

        listbox = Gtk.ListBox(activate_on_single_click=False)
        listbox.connect("row-activated", self.effects_listbox_row_activated_cb)

        expander.add(listbox)

        return expander

    def _create_effect_widget(self, effect_name, only_text):
        """Creates list box row for the given effect."""
        effect_info = self.app.effects.get_info(effect_name)

        effect_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, margin=SPACING / 2)
        effect_box.effect_name = effect_name
        effect_box.set_tooltip_text(effect_info.description)
        label = Gtk.Label(effect_info.human_name, xalign=0)

        if not only_text:
            # Show effect thumbnail
            icon = Gtk.Image.new_from_pixbuf(effect_info.icon)
            effect_box.pack_start(icon, False, True, SPACING / 2)

            # Set up favourite button
            fav_button = Gtk.Button()
            fav_button.props.relief = Gtk.ReliefStyle.NONE
            fav_button.props.halign = Gtk.Align.CENTER
            fav_button.props.valign = Gtk.Align.CENTER
            fav_button.set_tooltip_text(_("Add to Favourites"))

            starred = effect_name in self.app.settings.favourite_effects
            self._set_fav_button_state(fav_button, starred)
            fav_button.connect("clicked", self._fav_button_clicked_cb, effect_box.effect_name)
            effect_box.pack_end(fav_button, False, True, SPACING / 2)

        effect_box.pack_start(label, True, True, 0)

        # Set up drag behavoir
        eventbox = Gtk.EventBox(visible_window=False)
        eventbox.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, [EFFECT_TARGET_ENTRY], Gdk.DragAction.COPY)
        eventbox.connect("drag-data-get", self._drag_data_get_cb)
        eventbox.connect("drag-begin", self._drag_begin_cb)
        eventbox.add(effect_box)

        row = Gtk.ListBoxRow(selectable=False)
        row.add(eventbox)
        row.show_all()

        return row

    def _drag_data_get_cb(self, eventbox, drag_context, selection_data, unused_info, unused_timestamp):
        effect_box = eventbox.get_child()
        data = bytes(effect_box.effect_name, "UTF-8")
        selection_data.set(drag_context.list_targets()[0], 0, data)

    def _drag_begin_cb(self, eventbox, context):
        # Draw drag-icon
        icon = self._drag_icon
        icon_height = icon.get_height()
        icon_width = icon.get_width()

        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, icon_width, icon_height)
        ctx = cairo.Context(surface)
        # Center the icon around the cursor.
        ctx.translate(icon_width / 2, icon_height / 2)
        surface.set_device_offset(-icon_width / 2, -icon_height / 2)

        Gdk.cairo_set_source_pixbuf(ctx, icon, 0, 0)
        ctx.paint_with_alpha(0.35)

        Gtk.drag_set_icon_surface(context, surface)

    def effects_listbox_row_activated_cb(self, listbox, row):
        """Handles the activation of a row representing an effect."""
        effect_box = row.get_child().get_child()
        self.apply_effect(effect_box.effect_name)

    def apply_effect(self, effect_name):
        """Adds the selected effect to the single selected clip, if any."""
        effect_info = self.app.effects.get_info(effect_name)
        if not effect_info:
            return

        timeline = self.app.gui.editor.timeline_ui.timeline
        clip = timeline.selection.get_single_clip()
        if not clip:
            return

        pipeline = timeline.ges_timeline.get_parent()
        from pitivi.undo.timeline import CommitTimelineFinalizingAction
        with self.app.action_log.started("add effect",
                                         finalizing_action=CommitTimelineFinalizingAction(pipeline),
                                         toplevel=True):
            clip.ui.add_effect(effect_info)

    def _set_fav_button_state(self, button, is_active):
        """Manages the state of the favourite button."""
        button.active = is_active

        if button.active:
            image = Gtk.Image.new_from_pixbuf(self._star_icon_solid)
        else:
            image = Gtk.Image.new_from_pixbuf(self._star_icon_regular)

        button.props.image = image

    def _fav_button_clicked_cb(self, clicked_button, effect):
        """Adds effect to favourites and syncs the state of favourite button."""
        # Toggle the state of clicked button
        self._set_fav_button_state(clicked_button, not clicked_button.active)

        # Get all listboxes which contain the effect
        effect_info = self.app.effects.get_info(effect)
        all_effect_listboxes = [category_expander.get_child()
                                for category_expander in self.category_view.get_children()
                                if category_expander.get_label() in effect_info.categories]
        all_effect_listboxes.append(self.search_view)

        # Find and sync state in other listboxes
        for listbox in all_effect_listboxes:
            for row in listbox.get_children():
                effect_box = row.get_child().get_child()
                if effect == effect_box.effect_name:
                    fav_button = effect_box.get_children()[2]
                    # Sync the state with the clicked button
                    self._set_fav_button_state(fav_button, clicked_button.active)

        # Update the favourites list
        if clicked_button.active:
            self.app.settings.favourite_effects.append(effect)
        else:
            self.app.settings.favourite_effects = \
                [fav for fav in self.app.settings.favourite_effects if fav != effect]
        self.search_view.invalidate_filter()

    def _favourites_filter(self, row):
        """Filters search_view to show favourites."""
        effect_box = row.get_child().get_child()
        effect_name = effect_box.effect_name

        return effect_name in self.app.settings.favourite_effects

    def _favourites_toggle_cb(self, toggle):
        """Manages the visiblity and filtering of Favourites in search_view."""
        if toggle.get_active():
            self.search_entry.set_text("")
            self.search_view.set_filter_func(self._favourites_filter)
            self.search_view.invalidate_filter()
            self._switch_to_view(self.search_view)
        else:
            self._switch_to_view(self.category_view)

    def _search_filter(self, row):
        """Filters search_view to show search results."""
        effect_box = row.get_child().get_child()
        label = effect_box.get_children()[1]

        label_text = label.get_text().lower()
        search_key = self.search_entry.get_text().lower()

        return search_key in label_text

    def _search_entry_changed_cb(self, search_entry):
        """Manages the visiblity and filtering search results in search_view."""
        if search_entry.get_text():
            self.fav_view_toggle.props.active = False
            self.search_view.set_filter_func(self._search_filter)
            self.search_view.invalidate_filter()
            self._switch_to_view(self.search_view)
        else:
            self._switch_to_view(self.category_view)

    def _search_entry_icon_release_cb(self, entry, icon_pos, event):
        entry.set_text("")

    def _switch_to_view(self, next_view):
        """Shows next_view and hides all other views."""
        if next_view.props.visible:
            # It's already visible, no need to switch to it.
            return

        for child_view in self.main_view.get_children():
            child_view.props.visible = child_view == next_view


class EffectsPopover(Gtk.Popover, Loggable):
    """Popover for adding effects."""

    def __init__(self, app):
        Gtk.Popover.__init__(self)
        Loggable.__init__(self)

        self.app = app

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin=PADDING)

        self.search_entry = Gtk.SearchEntry()
        self.search_entry.connect("search-changed", self._search_entry_cb)

        scroll_window = Gtk.ScrolledWindow()
        scroll_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scroll_window.props.max_content_height = 350
        scroll_window.props.propagate_natural_height = True

        self.listbox = Gtk.ListBox()
        self.listbox.connect("row-activated", self._effect_row_activate_cb)
        self.listbox.set_filter_func(self._search_filter)
        placeholder_text = Gtk.Label(_("No effects"))
        placeholder_text.props.visible = True
        self.listbox.set_placeholder(placeholder_text)

        self.app.gui.editor.effectlist.add_effects_to_listbox(self.listbox, only_text=True)
        scroll_window.add(self.listbox)

        vbox.pack_start(self.search_entry, False, False, 0)
        vbox.pack_end(scroll_window, True, True, 0)
        vbox.show_all()

        self.add(vbox)

    def _effect_row_activate_cb(self, listbox, row):
        effect_box = row.get_child().get_child()
        self.app.gui.editor.effectlist.apply_effect(effect_box.effect_name)
        self.hide()

    def _search_entry_cb(self, search_entry):
        self.listbox.invalidate_filter()

    def _search_filter(self, row):
        effect_box = row.get_child().get_child()
        label = effect_box.get_children()[0]

        label_text = label.get_text().lower()
        search_key = self.search_entry.get_text().lower()

        return search_key in label_text

    def popup(self):
        self.search_entry.set_text("")
        Gtk.Popover.popup(self)


PROPS_TO_IGNORE = ['name', 'qos', 'silent', 'message', 'parent']


def create_widget_accumulator_func(ihint, return_accu, handler_return, *args):
    """Aborts `create_widget` emission if we got a widget."""
    continue_emission = handler_return is None
    return continue_emission, handler_return


class EffectsPropertiesManager(GObject.Object, Loggable):
    """Provides UIs for editing effects.

    Attributes:
        app (Pitivi): The app.
    """

    __gsignals__ = {
        "create_widget": (
            GObject.SignalFlags.RUN_LAST, Gtk.Widget, (GstElementSettingsWidget, GES.Effect,),
            create_widget_accumulator_func),
        "create_property_widget": (
            GObject.SignalFlags.RUN_LAST, object, (GstElementSettingsWidget, GES.Effect, object, object,),
            create_widget_accumulator_func),
    }

    def do_create_widget(self, effect_widget, effect):
        """Creates a widget if the `create_widget` handlers did not."""
        effect_name = effect.get_property("bin-description")
        self.log('UI is being auto-generated for "%s"', effect_name)
        effect_widget.add_widgets(create_property_widget_func=self.create_property_widget, with_reset_button=True)
        disable_scroll(effect_widget)
        self._post_configuration(effect, effect_widget)

    def do_create_property_widget(self, effect_widget, effect, prop, prop_value):
        """Creates a widget if the `create_property_widget` handlers did not."""
        widget = effect_widget.make_property_widget(prop, prop_value)
        return widget

    def __init__(self, app):
        GObject.Object.__init__(self)
        Loggable.__init__(self)
        self.app = app

    def get_effect_configuration_ui(self, effect):
        """Gets a configuration UI element for the effect.

        Args:
            effect (Gst.Element): The effect for which we want the UI.

        Returns:
            GstElementSettingsWidget: A container for configuring the effect.
        """
        effect_widget = GstElementSettingsWidget(effect, PROPS_TO_IGNORE)
        widget = self.emit("create_widget", effect_widget, effect)
        # The default handler of `create_widget` handles visibility
        # itself and returns None
        if widget is not None:
            effect_widget.show_widget(widget)
        self._connect_all_widget_callbacks(effect_widget, effect)

        return effect_widget

    def _post_configuration(self, effect, effect_set_ui):
        effect_name = effect.get_property("bin-description")
        if 'aspectratiocrop' in effect.get_property("bin-description"):
            for widget in effect_set_ui.get_children()[0].get_children():
                if isinstance(widget, FractionWidget):
                    widget.add_presets(["4:3", "5:4", "9:3", "16:9", "16:10"])
        else:
            self.log('No additional set-up required for "%s"', effect_name)
            return
        self.debug('Additional properties successfully set for "%s"', effect_name)

    def _connect_all_widget_callbacks(self, effect_widget, effect):
        for prop, widget in effect_widget.properties.items():
            widget.connect_value_changed(self._on_widget_value_changed_cb, widget, prop, effect, effect_widget)

    def _on_widget_value_changed_cb(self, unused_widget, prop_widget, prop, effect, effect_widget):
        if effect_widget.updating_property:
            # The widget is updated as a side-effect of setting one of its
            # properties. Ignore.
            return

        effect_widget.updating_property = True
        try:
            value = prop_widget.get_widget_value()

            # FIXME Workaround in order to make aspectratiocrop working
            if isinstance(value, Gst.Fraction):
                value = Gst.Fraction(int(value.num), int(value.denom))

            from pitivi.undo.timeline import CommitTimelineFinalizingAction
            pipeline = self.app.project_manager.current_project.pipeline
            with self.app.action_log.started("Effect property change",
                                             finalizing_action=CommitTimelineFinalizingAction(pipeline),
                                             toplevel=True):
                effect.set_child_property(prop.name, value)
        finally:
            effect_widget.updating_property = False

    def create_property_widget(self, element_settings_widget, prop, prop_value):
        prop_widget = self.emit("create_property_widget", element_settings_widget, element_settings_widget.element,
                                prop, prop_value)
        return prop_widget



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2013, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
from unittest import mock

from gi.repository import GES

from pitivi.utils.timeline import EditingContext
from pitivi.utils.timeline import SELECT
from pitivi.utils.timeline import SELECT_ADD
from pitivi.utils.timeline import Selected
from pitivi.utils.timeline import Selection
from pitivi.utils.timeline import UNSELECT
from tests import common


class TestSelected(common.TestCase):

    def test_bool_evaluation(self):
        selected = Selected()
        self.assertFalse(selected)

        selected.selected = True
        self.assertTrue(selected)

        selected.selected = False
        self.assertFalse(selected)


class TestSelection(common.TestCase):

    def test_bool_evaluation(self):
        clip1 = mock.MagicMock()
        selection = Selection()
        self.assertFalse(selection)
        selection.set_selection([clip1], SELECT)
        self.assertTrue(selection)
        selection.set_selection([clip1], SELECT_ADD)
        self.assertTrue(selection)
        selection.set_selection([clip1], UNSELECT)
        self.assertFalse(selection)

    def test_get_single_clip(self):
        selection = Selection()
        clip1 = common.create_test_clip(GES.UriClip)
        clip2 = common.create_test_clip(GES.TitleClip)

        # Selection empty.
        self.assertIsNone(selection.get_single_clip())
        self.assertIsNone(selection.get_single_clip(GES.UriClip))
        self.assertIsNone(selection.get_single_clip(GES.TitleClip))

        selection.set_selection([clip1], SELECT)
        self.assertEqual(selection.get_single_clip(), clip1)
        self.assertEqual(selection.get_single_clip(GES.UriClip), clip1)
        self.assertIsNone(selection.get_single_clip(GES.TitleClip))

        selection.set_selection([clip2], SELECT)
        self.assertEqual(selection.get_single_clip(), clip2)
        self.assertIsNone(selection.get_single_clip(GES.UriClip))
        self.assertEqual(selection.get_single_clip(GES.TitleClip), clip2)

        selection.set_selection([clip1, clip2], SELECT)
        self.assertIsNone(selection.get_single_clip())
        self.assertIsNone(selection.get_single_clip(GES.UriClip))
        self.assertIsNone(selection.get_single_clip(GES.TitleClip))

    def test_can_group_ungroup(self):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clip1, clip2 = self.add_clips_simple(timeline, 2)

        selection = Selection()
        self.assertFalse(selection.can_group)
        self.assertFalse(selection.can_ungroup)

        selection.set_selection([clip1], SELECT)
        self.assertFalse(selection.can_group)
        self.assertTrue(selection.can_ungroup)

        selection.set_selection([clip2], SELECT_ADD)
        self.assertTrue(selection.can_group)
        self.assertFalse(selection.can_ungroup)

        selection.set_selection([], SELECT)
        self.assertFalse(selection.can_group)
        self.assertFalse(selection.can_ungroup)

    def test_toplevels(self):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clip1, clip2, clip3, clip4 = self.add_clips_simple(timeline, 4)

        selection = Selection()

        selection.set_selection([clip1, clip2, clip3, clip4], SELECT)
        self.assertSetEqual(selection.toplevels, {clip1, clip2, clip3, clip4})

        group1 = GES.Container.group([clip1, clip2])
        group1.props.serialize = True
        self.assertSetEqual(selection.toplevels, {group1, clip3, clip4})

        group2 = GES.Container.group([group1, clip3])
        group2.props.serialize = True
        self.assertSetEqual(selection.toplevels, {group2, clip4})

        group1.props.serialize = True
        group1.props.serialize = False
        self.assertSetEqual(selection.toplevels, {group2, clip4})

        group1.props.serialize = False
        group2.props.serialize = False
        self.assertSetEqual(selection.toplevels, {clip1, clip2, clip3, clip4})

        group1.props.serialize = True
        group2.props.serialize = False
        self.assertSetEqual(selection.toplevels, {group1, clip3, clip4})


class TestEditingContext(common.TestCase):
    """Tests for the EditingContext class."""

    def test_with_video(self):
        """Checks the value of the with_video field."""
        audio_clip = common.get_sample_clip("mp3_sample.mp3")
        video_clip = common.get_sample_clip("one_fps_numeroted_blue.mkv")
        audio_video_clip = common.get_sample_clip("tears_of_steel.webm")

        # Add the clips to a layer so they have TrackElements.
        project = common.create_project()
        layer = project.ges_timeline.append_layer()
        self.assertTrue(layer.add_clip(audio_clip))
        self.assertTrue(layer.add_clip(video_clip))
        audio_video_clip.props.start = video_clip.props.duration
        self.assertTrue(layer.add_clip(audio_video_clip))

        self.__check_with_video(audio_clip, False)
        self.__check_with_video(video_clip, True)
        self.__check_with_video(audio_video_clip, True)

        # Check the track elements of a clip with audio only.
        audio_track_element = audio_clip.find_track_element(None, GES.AudioSource)
        self.__check_with_video(audio_track_element, False)

        # Check the track elements of a clip with video only.
        video_track_element = video_clip.find_track_element(None, GES.VideoSource)
        self.__check_with_video(video_track_element, True)

        # Check the track elements of a clip with both audio and video.
        audio_track_element = audio_video_clip.find_track_element(None, GES.AudioSource)
        video_track_element = audio_video_clip.find_track_element(None, GES.VideoSource)
        self.__check_with_video(audio_track_element, True)
        self.__check_with_video(video_track_element, True)

    def __check_with_video(self, clip, expected):
        context = EditingContext(clip, None, None, None, None, None)
        if expected:
            self.assertTrue(context.with_video)
        else:
            self.assertFalse(context.with_video)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2016 Jakub Brindza<jakub.brindza@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Accelerators info."""
import os.path

from gi.repository import GObject
from gi.repository import Gtk

from pitivi.settings import xdg_config_home
from pitivi.utils.misc import show_user_manual


class ShortcutsManager(GObject.Object):
    """Manager storing the shortcuts from all across the app."""

    __gsignals__ = {
        "accel-changed": (GObject.SignalFlags.RUN_LAST, None, (str,))
    }

    def __init__(self, app):
        GObject.Object.__init__(self)
        self.app = app
        self.__groups = []
        self.group_titles = {}
        self.group_actions = {}
        self.default_accelerators = {}
        self.titles = {}
        self.config_path = os.path.join(xdg_config_home(), "shortcuts.conf")
        self.__loaded_actions = list(self.__load())

    @property
    def groups(self):
        """The group ids ordered as they should be displayed."""
        return [item[1] for item in self.__groups]

    def __load(self):
        """Loads the shortcuts from the config file and sets them.

        Yields:
            string: The shortcuts loaded from the config file.
        """
        if not os.path.isfile(self.config_path):
            return

        for line in open(self.config_path, "r"):
            action_name, accelerators = line.split(":", 1)
            accelerators = accelerators.strip("\n").split(",")
            self.app.set_accels_for_action(action_name, accelerators)
            yield action_name

    def save(self):
        """Saves the accelerators for each action to the config file.

        Only the actions added using `add` with a title are considered.
        """
        with open(self.config_path, "w") as conf_file:
            for unused_group_id, actions in self.group_actions.items():
                for action, unused_title, unused_action_object in actions:
                    accels = ",".join(self.app.get_accels_for_action(action))
                    conf_file.write(action + ":" + accels + "\n")

    def add(self, action, accelerators, action_object, title, group=None):
        """Adds an action to be displayed.

        Args:
            action (str): The name identifying the action, formatted like
                "prefix.name".
            accelerators ([str]): The default accelerators corresponding to
                the action. They are set as the accelerators of the action
                only if no accelerators have been loaded from the config file
                initially, when the current manager instance has been created.
            action_object (Gio.SimpleAction): The object of the action.
            title (str): The title of the action.
            group (Optional[str]): The group id registered with `register_group`
                to be used instead of that extracted from `action`.
        """
        self.default_accelerators[action] = accelerators
        self.titles[action] = title
        if action not in self.__loaded_actions:
            self.app.set_accels_for_action(action, accelerators)

        action_prefix = group or action.split(".")[0]
        self.group_actions[action_prefix].append((action, title, action_object))

    def set(self, action, accelerators):
        """Sets accelerators for a shortcut.

        Args:
            action (str): The name identifying the action, formatted like
                "prefix.name".
            accelerators ([str]): The array containing accelerators to be set.
        """
        self.app.set_accels_for_action(action, accelerators)
        self.emit("accel-changed", action)

    def is_changed(self, action):
        """Checks whether the accelerators for an action have been changed.

        Args:
            action (str): The "prefix.name" identifying the action.

        Returns:
            bool: True iff the current accelerators are not the default ones.
        """
        accelerators = self.app.get_accels_for_action(action)
        return set(accelerators) != set(self.default_accelerators[action])

    def get_conflicting_action(self, action, keyval, mask):
        """Looks for a conflicting action using the specified accelerator.

        If an accelerator is used by another action in the same group or
        in the "win" and "app" global groups, it is not clear which of them
        will trigger when the accelerator is pressed.

        Args:
            action (str): The "prefix.name" identifying the action for which
                the accelerator will be set if there is no conflict.
            keyval (int): The key value of the accelerator.
            mask (int): The mask value of the accelerator.

        Returns:
            str: The name of the conflicting action using the accelerator, or None.
        """
        group_name = action.split(".")[0]
        for group in {group_name, "app", "win"}:
            for neighbor_action, unused_title, unused_action_object in self.group_actions[group]:
                if neighbor_action == action:
                    continue
                for accel in self.app.get_accels_for_action(neighbor_action):
                    if (keyval, mask) == Gtk.accelerator_parse(accel):
                        return neighbor_action
        return None

    def register_group(self, action_prefix, title, position):
        """Registers a group of shortcuts to be displayed.

        Args:
            action_prefix (str): The group id.
            title (str): The title of the group.
            position (int): The position used to sort the groups for display.
        """
        assert action_prefix not in self.group_titles
        self.group_titles[action_prefix] = title
        self.group_actions[action_prefix] = []
        self.__groups.append((position, action_prefix))
        self.__groups.sort()

    # pylint: disable=redefined-argument-from-local
    def reset_accels(self, action=None):
        """Resets accelerators to their default values.

        Args:
            action (Optional(str)): The action name.
                If specified, reset the specified action's accelerators.
                Otherwise reset accelerators for all actions.
        """
        if action:
            self.app.set_accels_for_action(action, self.default_accelerators[action])
            self.save()
            self.emit("accel-changed", action)
        else:
            for action, accelerators in self.default_accelerators.items():
                self.app.set_accels_for_action(action, accelerators)
            try:
                os.remove(self.config_path)
            except FileNotFoundError:
                pass
            self.emit("accel-changed", None)


class ShortcutsWindow(Gtk.ShortcutsWindow):
    """Dialog for displaying the accelerators."""

    def __init__(self, app):
        Gtk.ShortcutsWindow.__init__(self)
        self.app = app
        self.set_transient_for(self.app.gui)
        self.set_modal(True)
        self.populate()

    def populate(self):
        """Gathers the accelerators and populates the window."""
        section = Gtk.ShortcutsSection()
        section.show()
        for group_id in self.app.shortcuts.groups:
            actions = self.app.shortcuts.group_actions.get(group_id)
            if not actions:
                continue

            group = Gtk.ShortcutsGroup(title=self.app.shortcuts.group_titles[group_id])
            group.show()
            for action, title, _ in actions:
                # Show only the first accelerator which is the main one.
                # Don't bother with the others, to keep the dialog pretty.
                try:
                    accelerator = self.app.get_accels_for_action(action)[0]
                except IndexError:
                    accelerator = ""
                short = Gtk.ShortcutsShortcut(title=title, accelerator=accelerator)
                short.show()
                group.add(short)
            section.add(group)
        # Method below must be called after the section has been populated,
        # otherwise the shortcuts won't show up in search.
        self.add(section)


def show_shortcuts(app):
    """Shows the shortcuts window or the user manual page with the shortcuts."""
    if hasattr(Gtk, "ShortcutsWindow"):
        shortcuts_window = ShortcutsWindow(app)
        shortcuts_window.show()
    else:
        show_user_manual("cheatsheet")



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2020, Caleb Marcoux <caleb.marcoux@gmail.com> (primary contact),
# Copyright (c) 2020, Abby Brooks <abigail.brooks@huskers.unl.edu>,
# Copyright (c) 2020, Won Choi <won.choi@huskers.unl.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import json
import os

from gi.repository import GLib

from pitivi.settings import xdg_config_home
from pitivi.utils.loggable import Loggable


class EditorState(Loggable):
    """Pitivi editor state.

    Loads the editor state for the current project from the editor state folder.

    Widgets call get_value when they are ready to restore their state.
    """

    def __init__(self, project_manager):
        Loggable.__init__(self)

        self.project_manager = project_manager
        self.conf_folder_path = xdg_config_home("editor_state")

        self.project = None
        self.conf_file_path = None
        self._editor_state = {}
        self.__state_saving_handle = 0

    def get_value(self, key):
        """Get a value from the loaded editor state."""
        return self._editor_state.get(key)

    def set_value(self, key, value):
        """Sets the given value in the EditorState."""
        self._editor_state[key] = value
        self.prepare_to_save()

    def prepare_to_save(self):
        if self.__state_saving_handle:
            GLib.source_remove(self.__state_saving_handle)
        self.__state_saving_handle = GLib.timeout_add(500, self._state_not_changing_anymore_cb, priority=GLib.PRIORITY_LOW)

    def _state_not_changing_anymore_cb(self):
        self.__state_saving_handle = 0
        self.save_editor_state()

    def set_project(self, project):
        self.project = project
        self.conf_file_path = os.path.join(self.conf_folder_path, self.project.get_project_id() + ".conf")
        self.load_editor_state()

    def save_editor_state(self):
        """Save the current editor state to the editor state file."""
        self.log("Editor state saving.")

        if self.conf_file_path:
            with open(self.conf_file_path, "w") as file:
                json.dump(self._editor_state, file)

    def load_editor_state(self):
        """Load an editor state file into the current editor state."""
        self.log("Loading state from file: %s", self.conf_file_path)
        try:
            with open(self.conf_file_path, "r") as file:
                try:
                    self._editor_state = json.load(file)
                except (json.decoder.JSONDecodeError, ValueError) as e:
                    self.warning("Editor state could not be read: %s", e)
        except FileNotFoundError:
            return



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Dialog for browsing through projects to open an existing project."""
from gettext import gettext as _

from gi.repository import GES
from gi.repository import Gtk


class BrowseProjectsDialog(Gtk.FileChooserDialog):
    """Displays the Gtk.FileChooserDialog for browsing projects.

    Attributes:
        app (Pitivi): The app.
    """

    def __init__(self, app):
        Gtk.FileChooserDialog.__init__(self)

        self.set_title(_("Open Project…"))
        self.set_transient_for(app.gui)
        self.set_action(Gtk.FileChooserAction.OPEN)

        self.add_buttons(_("Cancel"), Gtk.ResponseType.CANCEL,
                         _("Open"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_select_multiple(False)
        self.set_current_folder(app.settings.lastProjectFolder)
        formatter_assets = GES.list_assets(GES.Formatter)
        formatter_assets.sort(
            key=lambda x: - x.get_meta(GES.META_FORMATTER_RANK))
        for format_ in formatter_assets:
            if format_.get_meta(GES.META_FORMATTER_NAME) == "base-formatter":
                continue
            file_filter = Gtk.FileFilter()
            file_filter.set_name(format_.get_meta(GES.META_DESCRIPTION))
            extension = format_.get_meta(GES.META_FORMATTER_EXTENSION)
            file_filter.add_pattern("*{}".format(extension))
            self.add_filter(file_filter)
        default = Gtk.FileFilter()
        default.set_name(_("All supported formats"))
        default.add_custom(Gtk.FileFilterFlags.URI, self.__can_load_uri, None)
        self.add_filter(default)

    # pylint: disable=bare-except
    @staticmethod
    def __can_load_uri(filterinfo, unused_uri):
        try:
            return GES.Formatter.can_load_uri(filterinfo.uri)
        except:
            return False



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005-2009 Edward Hervey <bilboed@bilboed.com>
# Copyright (c) 2008-2009 Alessandro Decina <alessandro.d@gmail.com>
# Copyright (c) 2014 Alexandru Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import os
import time
from gettext import gettext as _

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import Gtk

from pitivi.configure import RELEASES_URL
from pitivi.configure import VERSION
from pitivi.effects import EffectsManager
from pitivi.mainwindow import MainWindow
from pitivi.pluginmanager import PluginManager
from pitivi.project import ProjectManager
from pitivi.settings import GlobalSettings
from pitivi.settings import xdg_cache_home
from pitivi.shortcuts import ShortcutsManager
from pitivi.shortcuts import show_shortcuts
from pitivi.undo.project import ProjectObserver
from pitivi.undo.undo import UndoableActionLog
from pitivi.utils import loggable
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import path_from_uri
from pitivi.utils.misc import quote_uri
from pitivi.utils.proxy import ProxyManager
from pitivi.utils.system import get_system
from pitivi.utils.threads import ThreadMaster
from pitivi.utils.timeline import Zoomable


class Pitivi(Gtk.Application, Loggable):
    """Hello world.

    Attributes:
        action_log (UndoableActionLog): The undo/redo log for the current project.
        effects (EffectsManager): The effects which can be applied to a clip.
        gui (MainWindow): The main window of the app.
        recent_manager (Gtk.RecentManager): Manages recently used projects.
        project_manager (ProjectManager): The holder of the current project.
        settings (GlobalSettings): The application-wide settings.
        system (pitivi.utils.system.System): The system running the app.
    """

    __gsignals__ = {
        "version-info-received": (GObject.SignalFlags.RUN_LAST, None, (object,))
    }

    def __init__(self):
        Gtk.Application.__init__(self,
                                 application_id="org.pitivi.Pitivi",
                                 flags=Gio.ApplicationFlags.NON_UNIQUE | Gio.ApplicationFlags.HANDLES_OPEN)
        Loggable.__init__(self)

        self.settings = None
        self.threads = None
        self.effects = None
        self.system = None
        self.project_manager = ProjectManager(self)

        self.action_log = None
        self.project_observer = None
        self._last_action_time = Gst.util_get_timestamp()

        self.gui = None
        self.recent_manager = Gtk.RecentManager.get_default()
        self.__inhibit_cookies = {}

        self._version_information = {}

        self._scenario_file = None
        self._first_action = True

        Zoomable.app = self
        self.shortcuts = ShortcutsManager(self)

    def write_action(self, action, **kwargs):
        if self._scenario_file is None:
            return

        if self._first_action:
            self._scenario_file.write(
                "description, seek=true, handles-states=true\n")
            self._first_action = False

        now = Gst.util_get_timestamp()
        if now - self._last_action_time > 0.05 * Gst.SECOND:
            # We need to make sure that the waiting time was more than 50 ms.
            st = Gst.Structure.new_empty("wait")
            st["duration"] = float((now - self._last_action_time) / Gst.SECOND)
            self._scenario_file.write(st.to_string())
            self._scenario_file.write("\n")
            self._last_action_time = now

        if not isinstance(action, Gst.Structure):
            structure = Gst.Structure.new_empty(action)

            for key, value in kwargs.items():
                key = key.replace("_", "-")
                structure[key] = value

            action = structure

        self._scenario_file.write(action.to_string())
        self._scenario_file.write("\n")
        self._scenario_file.flush()

    def do_startup(self):
        Gtk.Application.do_startup(self)

        # Init logging as early as possible so we can log startup code
        enable_color = os.environ.get('PITIVI_DEBUG_NO_COLOR', '0') not in ('', '1')
        # Let's show a human-readable Pitivi debug output by default, and only
        # show a crazy unreadable mess when surrounded by gst debug statements.
        enable_crack_output = "GST_DEBUG" in os.environ
        loggable.init('PITIVI_DEBUG', enable_color, enable_crack_output)

        self.info('starting up')
        self._setup()
        self._check_version()

    def _setup(self):
        # pylint: disable=attribute-defined-outside-init
        self.settings = GlobalSettings()
        self.threads = ThreadMaster()
        self.effects = EffectsManager()
        self.proxy_manager = ProxyManager(self)
        self.system = get_system()
        self.plugin_manager = PluginManager(self)

        self.project_manager.connect("new-project-loaded", self._new_project_loaded_cb)
        self.project_manager.connect_after("project-closed", self._project_closed_cb)
        self.project_manager.connect("project-saved", self.__project_saved_cb)

        self._create_actions()
        self._sync_do_undo()

    def _create_actions(self):
        self.shortcuts.register_group("app", _("General"), position=10)

        # pylint: disable=attribute-defined-outside-init
        self.undo_action = Gio.SimpleAction.new("undo", None)
        self.undo_action.connect("activate", self._undo_cb)
        self.add_action(self.undo_action)
        self.shortcuts.add("app.undo", ["<Primary>z"], self.undo_action,
                           _("Undo the most recent action"))

        self.redo_action = Gio.SimpleAction.new("redo", None)
        self.redo_action.connect("activate", self._redo_cb)
        self.add_action(self.redo_action)
        self.shortcuts.add("app.redo", ["<Primary><Shift>z"], self.redo_action,
                           _("Redo the most recent action"))

        self.quit_action = Gio.SimpleAction.new("quit", None)
        self.quit_action.connect("activate", self._quit_cb)
        self.add_action(self.quit_action)
        self.shortcuts.add("app.quit", ["<Primary>q"], self.quit_action, _("Quit"))

        self.show_shortcuts_action = Gio.SimpleAction.new("shortcuts_window", None)
        self.show_shortcuts_action.connect("activate", self._show_shortcuts_cb)
        self.add_action(self.show_shortcuts_action)
        self.shortcuts.add("app.shortcuts_window",
                           ["<Primary>F1", "<Primary>question"],
                           self.show_shortcuts_action,
                           _("Show the Shortcuts Window"))

    def do_activate(self):
        if self.gui:
            # The app is already started and the window already created.
            # Present the already existing window.
            if self.system.has_x11():
                # TODO: Use present() instead of present_with_time() when
                # https://bugzilla.gnome.org/show_bug.cgi?id=688830 is fixed.
                from gi.repository import GdkX11
                x11_server_time = GdkX11.x11_get_server_time(self.gui.get_window())
                self.gui.present_with_time(x11_server_time)
            else:
                # On Wayland or Quartz (Mac OS X) backend there is no GdkX11,
                # so just use present() directly here.
                self.gui.present()
            # No need to show the welcome wizard.
            return

        self.create_main_window()
        self.gui.show_perspective(self.gui.greeter)

    def create_main_window(self):
        if self.gui:
            return
        self.gui = MainWindow(self)
        self.gui.setup_ui()
        self.add_window(self.gui)

    def do_open(self, giofiles, unused_count, unused_hint):
        assert giofiles
        self.create_main_window()
        if len(giofiles) > 1:
            self.warning("Opening only one project at a time. Ignoring the rest!")
        project_file = giofiles[0]
        self.project_manager.load_project(quote_uri(project_file.get_uri()))
        return True

    def shutdown(self):
        """Closes the app.

        Returns:
            bool: True if successful, False otherwise.
        """
        self.debug("shutting down")
        # Refuse to close if we are not done with the current project.
        if not self.project_manager.close_running_project():
            self.warning(
                "Not closing since running project doesn't want to close")
            return False
        if self.gui:
            self.gui.destroy()
        self.threads.wait_all_threads()
        self.settings.store_settings()
        self.quit()
        return True

    def _set_scenario_file(self, uri):
        if uri:
            project_path = path_from_uri(uri)
        else:
            # New project.
            project_path = None
        if 'PITIVI_SCENARIO_FILE' in os.environ:
            scenario_path = os.environ['PITIVI_SCENARIO_FILE']
        else:
            cache_dir = xdg_cache_home("scenarios")
            scenario_name = str(time.strftime("%Y%m%d-%H%M%S"))
            if project_path:
                scenario_name += os.path.splitext(project_path.replace(os.sep, "_"))[0]
            scenario_path = os.path.join(cache_dir, scenario_name + ".scenario")

        scenario_path = path_from_uri(quote_uri(scenario_path))
        self._scenario_file = open(scenario_path, "w")

        if project_path and not project_path.endswith(".scenario"):
            # It's an xges file probably.
            with open(project_path) as project:
                content = project.read().replace("\n", "")
                self.write_action("load-project",
                                  serialized_content=content)

    def _new_project_loaded_cb(self, unused_project_manager, project):
        uri = project.get_uri()
        if uri:
            # We remove the project from recent projects list
            # and then re-add it to this list to make sure it
            # gets positioned at the top of the recent projects list.
            try:
                self.recent_manager.remove_item(uri)
            except GLib.Error as e:
                if e.domain != "gtk-recent-manager-error-quark":
                    raise e
            self.recent_manager.add_item(uri)

        self.action_log = UndoableActionLog()
        self.action_log.connect("pre-push", self._action_log_pre_push_cb)
        self.action_log.connect("commit", self._action_log_commit)
        self.action_log.connect("move", self._action_log_move_cb)
        self.project_observer = ProjectObserver(project, self.action_log)

        self._set_scenario_file(project.get_uri())

    def __project_saved_cb(self, unused_project_manager, unused_project, uri):
        if uri:
            self.recent_manager.add_item(uri)

    def _project_closed_cb(self, unused_project_manager, project):
        if project.loaded:
            self.action_log = None
            self._sync_do_undo()

        if self._scenario_file:
            self.write_action("stop")
            self._scenario_file.close()
            self._scenario_file = None

    def _check_version(self):
        """Checks online for new versions of the app."""
        self.info("Requesting version information async")
        giofile = Gio.File.new_for_uri(RELEASES_URL)
        giofile.load_contents_async(None, self._version_info_received_cb, None)

    def _version_info_received_cb(self, giofile, result, user_data):
        # pylint: disable=broad-except
        try:
            raw = giofile.load_contents_finish(result)[1]
            if not isinstance(raw, str):
                raw = raw.decode()
            raw = raw.split("\n")
            # Split line at '=' if the line is not empty or a comment line
            data = [element.split("=") for element in raw
                    if element and not element.startswith("#")]

            # search newest version and status
            status = "UNSUPPORTED"
            current_version = None
            for version, version_status in data:
                if VERSION == version:
                    status = version_status
                if version_status.upper() == "CURRENT":
                    # This is the latest.
                    current_version = version
                    self.info("Latest software version is %s", current_version)

            version_split = [int(i) for i in VERSION.split(".")]
            current_version_split = [int(i)
                                     for i in current_version.split(".")]
            if version_split > current_version_split:
                status = "CURRENT"
                self.info(
                    "Running version %s, which is newer than the latest known version. Considering it as the latest current version.", VERSION)
            elif status == "UNSUPPORTED":
                self.warning(
                    "Using an outdated version of Pitivi (%s)", VERSION)

            self._version_information["current"] = current_version
            self._version_information["status"] = status
            self.emit("version-info-received", self._version_information)
        except Exception as e:
            self.warning("Version info could not be read: %s", e)

    def is_latest(self):
        """Whether the app's version is the latest as far as we know."""
        status = self._version_information.get("status")
        return status is None or status.upper() == "CURRENT"

    def get_latest(self):
        """Get the latest version of the app or None."""
        return self._version_information.get("current")

    def _quit_cb(self, unused_action, unused_param):
        self.shutdown()

    def _undo_cb(self, unused_action, unused_param):
        self.action_log.undo()

    def _redo_cb(self, unused_action, unused_param):
        self.action_log.redo()

    def _show_shortcuts_cb(self, unused_action, unused_param):
        show_shortcuts(self)

    def _action_log_pre_push_cb(self, unused_action_log, action):
        scenario_action = action.as_scenario_action()
        if scenario_action:
            self.write_action(scenario_action)

    def _action_log_commit(self, action_log, unused_stack):
        if action_log.is_in_transaction():
            return
        self._sync_do_undo()

    def _action_log_move_cb(self, action_log, unused_stack):
        self._sync_do_undo()

    def _sync_do_undo(self):
        can_undo = self.action_log and bool(self.action_log.undo_stacks)
        self.undo_action.set_enabled(bool(can_undo))

        can_redo = self.action_log and bool(self.action_log.redo_stacks)
        self.redo_action.set_enabled(bool(can_redo))

        if not self.project_manager.current_project:
            return

        dirty = self.action_log and self.action_log.dirty()
        self.project_manager.current_project.set_modification_state(dirty)
        # In the tests we do not want to create any gui
        if self.gui is not None:
            self.gui.editor.show_project_status()

    def simple_inhibit(self, reason, flags):
        """Informs the session manager about actions to be inhibited.

        Keeps track of the reasons received. A specific reason should always
        be accompanied by the same flags. Calling the method a second time
        with the same reason has no effect unless `simple_uninhibit` has been
        called in the meanwhile.

        Args:
            reason (str): The reason for which to perform the inhibition.
            flags (Gtk.ApplicationInhibitFlags): What should be inhibited.
        """
        if reason in self.__inhibit_cookies:
            self.debug("Inhibit reason already active: %s", reason)
            return
        self.debug("Inhibiting %s for %s", flags, reason)
        cookie = self.inhibit(self.gui, flags, reason)
        self.__inhibit_cookies[reason] = cookie

    def simple_uninhibit(self, reason):
        """Informs the session manager that an inhibition is not needed anymore.

        Args:
            reason (str): The reason which is not valid anymore.
        """
        try:
            cookie = self.__inhibit_cookies.pop(reason)
        except KeyError:
            self.debug("Inhibit reason not active: %s", reason)
            return
        self.debug("Uninhibiting %s", reason)
        self.uninhibit(cookie)
